"""Tests for ecephys_job module."""

import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from aind_data_transfer.transformations.ephys_compressors import (
    EcephysCompressionParameters,
    EphysCompressors,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"
DATA_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream"
BEHAVIOR_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream" / "Videos"


class TestEcephysCompression(unittest.TestCase):
    """Tests for EcephysJob class"""

    @patch("shutil.copytree")
    @patch("shutil.ignore_patterns")
    @patch("aind_data_transfer.transformations.ephys_compressors.memmap")
    @patch(
        "aind_data_transfer.readers.ephys_readers.EphysReaders."
        "get_streams_to_clip"
    )
    @patch(
        "aind_data_transfer.readers.ephys_readers.EphysReaders."
        "get_read_blocks"
    )
    @patch(
        "aind_data_transfer.transformations.ephys_compressors."
        "EphysCompressors.get_compressor"
    )
    @patch(
        "aind_data_transfer.transformations.ephys_compressors."
        "EphysCompressors.scale_read_blocks"
    )
    @patch(
        "aind_data_transfer.writers.ephys_writers.EphysWriters."
        "compress_and_write_block"
    )
    @patch(
        "aind_data_transfer.transformations.ephys_compressors."
        "correct_np_opto_electrode_locations"
    )
    def test_ecephys_job_with_compression(
        self,
        mock_correct_np_opto: MagicMock,
        mock_write_block: MagicMock,
        mock_scale_read_blocks: MagicMock,
        mock_get_compressor: MagicMock,
        mock_get_read_blocks: MagicMock,
        mock_get_streams_to_clip: MagicMock,
        mock_memmap: MagicMock,
        mock_ignore_patterns: MagicMock,
        mock_copytree: MagicMock,
    ):
        """Tests ecephys job runs correctly with compression"""
        mock_get_streams_to_clip.return_value.__iter__.return_value = [
            {"data": [], "relative_path_name": "some_rel_path", "n_chan": 1}
        ]
        mock_ignore_patterns.side_effect = [
            ["*.dat"],
            ["*.dat", str(BEHAVIOR_DIR / "*")],
        ]
        mock_get_compressor.return_value = "Mocked Compressor"
        mock_get_read_blocks.return_value = {
            "recording": "mocked recording",
            "experiment_name": "mocked exp name",
            "stream_name": "mocked stream name",
        }
        mock_scale_read_blocks.return_value = {
            "scaled_recording": "mocked scale rec",
            "experiment_name": "mocked exp name",
            "stream_name": "mocked stream name",
        }

        ecephys_configs = EcephysCompressionParameters(source=DATA_DIR)
        ecephys_compressor = EphysCompressors(
            job_configs=ecephys_configs, behavior_dir=BEHAVIOR_DIR
        )
        ecephys_compressor.compress_raw_data(temp_dir=Path("some_path"))
        ecephys_compressor.compress_raw_data(temp_dir=Path("some_path"))
        mock_correct_np_opto.assert_has_calls([call(DATA_DIR), call(DATA_DIR)])
        mock_copytree.assert_has_calls(
            [
                call(
                    DATA_DIR,
                    Path("some_path/ecephys_clipped"),
                    ignore=["*.dat"],
                ),
                call(
                    DATA_DIR,
                    Path("some_path/ecephys_clipped"),
                    ignore=["*.dat", str(BEHAVIOR_DIR / "*")],
                ),
            ]
        )
        mock_get_streams_to_clip.assert_has_calls(
            [
                call("openephys", DATA_DIR),
                call().__iter__(),
                call("openephys", DATA_DIR),
                call().__iter__(),
            ]
        )
        mock_memmap.assert_has_calls(
            [
                call(
                    Path("some_path/ecephys_clipped/some_rel_path"),
                    dtype="int16",
                    shape=(100, 1),
                    order="C",
                    mode="w+",
                ),
                call().__setitem__(slice(None, None, None), []),
                call(
                    Path("some_path/ecephys_clipped/some_rel_path"),
                    dtype="int16",
                    shape=(100, 1),
                    order="C",
                    mode="w+",
                ),
                call().__setitem__(slice(None, None, None), []),
            ]
        )
        mock_get_read_blocks.assert_has_calls(
            [call("openephys", DATA_DIR), call("openephys", DATA_DIR)]
        )

        mock_get_compressor.assert_has_calls(
            [call("wavpack", level=3), call("wavpack", level=3)]
        )
        mock_scale_read_blocks.assert_has_calls(
            [
                call(
                    read_blocks={
                        "recording": "mocked recording",
                        "experiment_name": "mocked exp name",
                        "stream_name": "mocked stream name",
                    },
                    num_chunks_per_segment=100,
                    chunk_size=10000,
                ),
                call(
                    read_blocks={
                        "recording": "mocked recording",
                        "experiment_name": "mocked exp name",
                        "stream_name": "mocked stream name",
                    },
                    num_chunks_per_segment=100,
                    chunk_size=10000,
                ),
            ]
        )
        mock_write_block.assert_has_calls(
            [
                call(
                    read_blocks={
                        "scaled_recording": "mocked scale rec",
                        "experiment_name": "mocked exp name",
                        "stream_name": "mocked stream name",
                    },
                    compressor="Mocked Compressor",
                    output_dir=Path("some_path/ecephys_compressed"),
                    max_windows_filename_len=150,
                    output_format="zarr",
                    job_kwargs={"n_jobs": -1},
                ),
                call(
                    read_blocks={
                        "scaled_recording": "mocked scale rec",
                        "experiment_name": "mocked exp name",
                        "stream_name": "mocked stream name",
                    },
                    compressor="Mocked Compressor",
                    output_dir=Path("some_path/ecephys_compressed"),
                    max_windows_filename_len=150,
                    output_format="zarr",
                    job_kwargs={"n_jobs": -1},
                ),
            ]
        )


if __name__ == "__main__":
    unittest.main()



import logging
import platform
import subprocess

import boto3
from botocore.exceptions import ClientError


def get_secret(secret_name, region_name):
    """Retrieves a secret"""

    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(
        service_name="secretsmanager", region_name=region_name
    )

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        # For a list of exceptions thrown, see
        # https://docs.aws.amazon.com/secretsmanager/latest/apireference/
        #   API_GetSecretValue.html
        raise e

    # Decrypts secret using the associated KMS key.
    # Don't log or print!
    secret = get_secret_value_response["SecretString"]

    return secret


def upload_to_s3(
    directory_to_upload,
    s3_bucket,
    s3_prefix,
    dryrun=False,
    excluded=None,
    included=None,
):
    # Upload to s3
    if platform.system() == "Windows":
        shell = True
    else:
        shell = False

    # TODO: Use s3transfer library instead of subprocess?
    logging.info("Uploading to s3.")
    aws_dest = f"s3://{s3_bucket}/{s3_prefix}"
    base_command = ["aws", "s3", "sync", directory_to_upload, aws_dest]

    if excluded:
        base_command.extend(["--exclude", excluded])
    if included:
        base_command.extend(["--include", included])
    if dryrun:
        base_command.append("--dryrun")
    subprocess.run(base_command, shell=shell, check=True)

    logging.info("Finished uploading to s3.")


def copy_to_s3(file_to_upload, s3_bucket, s3_prefix, dryrun):
    # Upload to s3
    if platform.system() == "Windows":
        shell = True
    else:
        shell = False

    # TODO: Use s3transfer library instead of subprocess?
    logging.info("Copying file to s3.")
    aws_dest = f"s3://{s3_bucket}/{s3_prefix}"
    if dryrun:
        subprocess.run(
            [
                "aws",
                "s3",
                "cp",
                file_to_upload,
                aws_dest,
                "--dryrun",
            ],
            shell=shell,
            check=True,
        )
    else:
        subprocess.run(
            ["aws", "s3", "cp", file_to_upload, aws_dest],
            shell=shell,
            check=True,
        )
    logging.info("Finished copying file to s3.")



"""Package to compress and upload data to the cloud
"""

import logging

LOG_FMT = "%(asctime)s %(message)s"
LOG_DATE_FMT = "%Y-%m-%d %H:%M"

logging.basicConfig(format=LOG_FMT, datefmt=LOG_DATE_FMT)

__version__ = "0.21.4"



"""Tests the basic_job module"""

import json
import os
import unittest
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from requests import Response

from aind_data_transfer import __version__
from aind_data_transfer.config_loader.base_config import BasicUploadJobConfigs
from aind_data_transfer.jobs.basic_job import BasicJob
from aind_data_transfer.transformations.metadata_creation import (
    ProceduresMetadata,
    SubjectMetadata,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"
DATA_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream"
CONFIG_FILE = TEST_DIR / "test_configs" / "ephys_upload_job_test_configs.yml"
BEHAVIOR_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream" / "Videos"
METADATA_DIR = TEST_DIR / "test_metadata"

with open(METADATA_DIR / "processing.json", "r") as f:
    example_processing_instance_json = json.load(f)

with open(METADATA_DIR / "data_description.json", "r") as f:
    example_data_description_instance_json = json.load(f)

with open(METADATA_DIR / "subject.json", "r") as f:
    example_subject_instance_json = json.load(f)

with open(METADATA_DIR / "procedures.json", "r") as f:
    example_procedures_instance_json = json.load(f)


class TestBasicJob(unittest.TestCase):
    """Tests methods in the BasicJob class"""

    EXAMPLE_ENV_VAR1 = {
        "CODEOCEAN_DOMAIN": "some_domain",
        "CODEOCEAN_TRIGGER_CAPSULE_ID": "some_capsule_id",
        "METADATA_SERVICE_DOMAIN": "some_ms_domain",
        "AIND_DATA_TRANSFER_REPO_LOCATION": "some_dtr_location",
        "VIDEO_ENCRYPTION_PASSWORD": "some_password",
        "CODEOCEAN_API_TOKEN": "some_api_token",
        "S3_BUCKET": "some_bucket",
        "MODALITIES": f'[{{"modality":"MRI",' f'"source":"{str(DATA_DIR)}"}}]',
        "EXPERIMENT_TYPE": "confocal",
        "SUBJECT_ID": "12345",
        "ACQ_DATE": "2020-10-10",
        "ACQ_TIME": "10:10:10",
        "DATA_SOURCE": str(DATA_DIR),
        "DRY_RUN": "true",
    }

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("tempfile.TemporaryDirectory")
    @patch("aind_data_transfer.jobs.basic_job.upload_to_s3")
    def test_aws_creds_check_allowed(
        self,
        mock_upload_to_s3: MagicMock,
        mock_tempfile: MagicMock,
    ):
        """Tests that the aws credentials pass allowed"""
        mock_tempfile.return_value.__enter__.return_value = (
            Path("some_dir") / "tmp"
        )
        basic_job_configs = BasicUploadJobConfigs()
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._test_upload(Path("some_dir"))
        mock_upload_to_s3.assert_called_once_with(
            directory_to_upload=Path("some_dir"),
            s3_bucket="some_bucket",
            s3_prefix="confocal_12345_2020-10-10_10-10-10",
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("os.mkdir")
    @patch(
        "aind_data_transfer.transformations.generic_compressors.ZipCompressor."
        "compress_dir"
    )
    @patch("shutil.copytree")
    def test_compress_raw_data_no_zip(
        self,
        mock_copytree: MagicMock,
        mock_compress: MagicMock,
        mock_make_dir: MagicMock,
    ):
        """Tests that the raw data is copied to temp directory"""
        basic_job_configs = BasicUploadJobConfigs()
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._compress_raw_data(temp_dir=Path("some_path"))

        # The shutil copy takes care of creating the directory
        mock_make_dir.assert_not_called()
        # With a Behavior directory defined
        basic_job.job_configs.behavior_dir = BEHAVIOR_DIR
        basic_job._compress_raw_data(temp_dir=Path("some_path"))
        mock_copytree.assert_has_calls(
            [
                call(DATA_DIR, Path("some_path/mri"), ignore=None),
                call(
                    DATA_DIR,
                    Path("some_path/mri"),
                    ignore=(BEHAVIOR_DIR / "*"),
                ),
            ]
        )

        self.assertFalse(mock_compress.called)

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch(
        "aind_data_transfer.transformations.generic_compressors.ZipCompressor."
        "compress_dir"
    )
    @patch("os.mkdir")
    def test_compress_raw_data_with_zip(
        self, mock_mkdir: MagicMock, mock_compress: MagicMock
    ):
        """Tests that the raw data is zipped"""
        basic_job_configs = BasicUploadJobConfigs()
        basic_job_configs.modalities[0].compress_raw_data = True
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._compress_raw_data(temp_dir=Path("some_path"))

        # With a behavior directory defined
        basic_job_configs.behavior_dir = BEHAVIOR_DIR
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._compress_raw_data(temp_dir=Path("some_path"))

        mock_mkdir.assert_has_calls(
            [
                call(Path("some_path/mri")),
                call(Path("some_path/mri")),
            ]
        )

        mock_compress.assert_has_calls(
            [
                call(
                    input_dir=DATA_DIR,
                    output_dir=Path(
                        "some_path/mri/"
                        "v0.6.x_neuropixels_multiexp_multistream.zip"
                    ),
                    skip_dirs=None,
                ),
                call(
                    input_dir=DATA_DIR,
                    output_dir=Path(
                        "some_path/mri/"
                        "v0.6.x_neuropixels_multiexp_multistream.zip"
                    ),
                    skip_dirs=[BEHAVIOR_DIR],
                ),
            ]
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch(
        "aind_data_transfer.transformations.metadata_creation."
        "SubjectMetadata.from_service"
    )
    @patch(
        "aind_data_transfer.transformations.metadata_creation."
        "ProceduresMetadata.from_service"
    )
    @patch(
        "aind_data_transfer.transformations.metadata_creation."
        "MetadataCreation.write_to_json"
    )
    @patch("os.path.exists")
    @patch("shutil.copyfile")
    @patch("aind_data_transfer.jobs.basic_job.datetime")
    def test_compile_metadata(
        self,
        mock_datetime: MagicMock,
        mock_copyfile: MagicMock,
        mock_path_exists: MagicMock,
        mock_json_write: MagicMock,
        mock_procedures_service: MagicMock,
        mock_subject_service: MagicMock,
    ):
        """Tests that the metadata files are compiled correctly."""

        mock_datetime.now.return_value = datetime(2023, 4, 9)
        mock_path_exists.return_value = True
        mock_subject_service.return_value = SubjectMetadata(
            model_obj=example_subject_instance_json
        )
        mock_procedures_service.return_value = ProceduresMetadata(
            model_obj=example_procedures_instance_json
        )

        basic_job_configs = BasicUploadJobConfigs()
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._compile_metadata(
            Path("some_dir"), process_start_time=datetime(2023, 4, 9)
        )

        # Test where metadata directory is defined
        basic_job_configs.metadata_dir = METADATA_DIR
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._compile_metadata(
            Path("some_dir"), process_start_time=datetime(2023, 4, 9)
        )

        # Test where metadata dir forced is true
        basic_job_configs.metadata_dir_force = True
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._compile_metadata(
            Path("some_dir"), process_start_time=datetime(2023, 4, 9)
        )

        mock_json_write.assert_has_calls(
            [
                call(Path("some_dir/subject.json")),
                call(Path("some_dir/procedures.json")),
                call(Path("some_dir/data_description.json")),
                call(Path("some_dir/processing.json")),
            ],
            any_order=True,
        )

        mock_copyfile.assert_has_calls(
            [
                call(
                    str(METADATA_DIR / "data_description.json"),
                    Path("some_dir/data_description.json"),
                ),
                call(
                    str(METADATA_DIR / "processing.json"),
                    Path("some_dir/processing.json"),
                ),
                call(
                    str(METADATA_DIR / "subject.json"),
                    Path("some_dir/subject.json"),
                ),
                call(
                    str(METADATA_DIR / "procedures.json"),
                    Path("some_dir/procedures.json"),
                ),
            ],
            any_order=True,
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("os.makedirs")
    @patch("shutil.copy")
    @patch(
        "aind_data_transfer.transformations.generic_compressors."
        "VideoCompressor.compress_all_videos_in_dir"
    )
    def test_encrypt_behavior_dir(
        self,
        mock_compress: MagicMock,
        mock_copy: MagicMock,
        mock_mkdirs: MagicMock,
    ):
        """Tests that the video files in the behavior dir are encrypted."""
        basic_job_configs = BasicUploadJobConfigs()
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job_configs.behavior_dir = BEHAVIOR_DIR
        basic_job._encrypt_behavior_dir(temp_dir=Path("some_dir"))
        mock_copy.assert_has_calls(
            [
                call(
                    str(BEHAVIOR_DIR / "a_video_file.avi"),
                    "some_dir/behavior/a_video_file.avi",
                ),
                call(
                    str(BEHAVIOR_DIR / "another_video_file.avi"),
                    "some_dir/behavior/another_video_file.avi",
                ),
                call(
                    str(BEHAVIOR_DIR / "video_ref.csv"),
                    "some_dir/behavior/video_ref.csv",
                ),
            ],
            any_order=True,
        )
        mock_compress.assert_called_once_with(Path("some_dir/behavior"))
        mock_mkdirs.assert_has_calls(
            [
                call("some_dir/behavior", exist_ok=True),
                call("some_dir/behavior", exist_ok=True),
                call("some_dir/behavior", exist_ok=True),
            ]
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("aind_data_transfer.jobs.basic_job.upload_to_s3")
    def test_upload_to_s3(
        self,
        mock_upload: MagicMock,
    ):
        """Tests that the data is uploaded to S3"""
        basic_job_configs = BasicUploadJobConfigs()
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._upload_to_s3(temp_dir=Path("some_dir"))
        mock_upload.assert_called_once_with(
            directory_to_upload=Path("some_dir"),
            s3_bucket="some_bucket",
            s3_prefix="confocal_12345_2020-10-10_10-10-10",
            dryrun=True,
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("aind_codeocean_api.codeocean.CodeOceanClient.run_capsule")
    def test_trigger_codeocean_capsule(
        self,
        mock_run_capsule: MagicMock,
    ):
        """Tests code ocean capsule is triggered"""
        successful_response = Response()
        successful_response.status_code = 200
        successful_response._content = json.dumps(
            {"Message": "triggered a code ocean capsule"}
        ).encode("utf-8")
        mock_run_capsule.return_value = successful_response
        # With dry-run set to True
        basic_job_configs = BasicUploadJobConfigs()
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._trigger_codeocean_pipeline()

        # With dry-run to False
        basic_job.job_configs.dry_run = False
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._trigger_codeocean_pipeline()

        mock_run_capsule.assert_called_once_with(
            capsule_id="some_capsule_id",
            data_assets=[],
            parameters=[
                '{"trigger_codeocean_job": '
                '{"job_type": "confocal", '
                '"modalities": ["MRI"], '
                '"capsule_id": "some_capsule_id", '
                '"process_capsule_id": null, '
                '"bucket": "some_bucket", '
                '"prefix": "confocal_12345_2020-10-10_10-10-10", '
                f'"aind_data_transfer_version": "{__version__}"'
                "}}"
            ],
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("aind_codeocean_api.codeocean.CodeOceanClient.run_capsule")
    def test_trigger_custom_codeocean_capsule(
        self,
        mock_run_capsule: MagicMock,
    ):
        """Tests code ocean capsule is triggered"""
        successful_response = Response()
        successful_response.status_code = 200
        successful_response._content = json.dumps(
            {"Message": "triggered a code ocean capsule"}
        ).encode("utf-8")
        mock_run_capsule.return_value = successful_response
        # With dry-run set to True
        basic_job_configs = BasicUploadJobConfigs()
        basic_job_configs.dry_run = False
        basic_job_configs.codeocean_process_capsule_id = "xyz-456"
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job._trigger_codeocean_pipeline()

        mock_run_capsule.assert_called_once_with(
            capsule_id="some_capsule_id",
            data_assets=[],
            parameters=[
                '{"trigger_codeocean_job": '
                '{"job_type": "run_generic_pipeline", '
                '"modalities": ["MRI"], '
                '"capsule_id": "some_capsule_id", '
                '"process_capsule_id": "xyz-456", '
                '"bucket": "some_bucket", '
                '"prefix": "confocal_12345_2020-10-10_10-10-10", '
                f'"aind_data_transfer_version": "{__version__}"'
                "}}"
            ],
        )

    @patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @patch("tempfile.TemporaryDirectory")
    @patch(
        "aind_data_transfer.jobs.basic_job.BasicJob."
        "_check_if_s3_location_exists"
    )
    @patch("aind_data_transfer.jobs.basic_job.BasicJob._compress_raw_data")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob._compile_metadata")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob._encrypt_behavior_dir")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob._upload_to_s3")
    @patch(
        "aind_data_transfer.jobs.basic_job.BasicJob."
        "_trigger_codeocean_pipeline"
    )
    @patch("aind_data_transfer.jobs.basic_job.datetime")
    @patch("aind_data_transfer.jobs.basic_job.BasicJob._test_upload")
    def test_run_job(
        self,
        mock_test_upload: MagicMock,
        mock_datetime: MagicMock,
        mock_trigger_pipeline: MagicMock,
        mock_upload_to_s3: MagicMock,
        mock_encrypt_behavior: MagicMock,
        mock_compile_metadata: MagicMock,
        mock_compress_raw_data: MagicMock,
        mock_s3_check: MagicMock,
        mock_tempfile: MagicMock,
    ):
        """Tests that the run_job executes all sub jobs"""

        mock_datetime.now.return_value = datetime(2023, 4, 9)

        mock_tempfile.return_value.__enter__.return_value = (
            Path("some_dir") / "tmp"
        )
        basic_job_configs = BasicUploadJobConfigs()
        # Set the temp directory
        basic_job_configs.temp_directory = "some_dir"
        basic_job = BasicJob(job_configs=basic_job_configs)
        basic_job.run_job()

        mock_test_upload.assert_called_once_with(
            temp_dir=(Path("some_dir") / "tmp")
        )
        mock_tempfile.assert_called_once_with(dir="some_dir")
        mock_s3_check.assert_called_once()
        mock_compress_raw_data.assert_called_once_with(
            temp_dir=(Path("some_dir") / "tmp")
        )
        mock_compile_metadata.assert_called_once_with(
            temp_dir=(Path("some_dir") / "tmp"),
            process_start_time=datetime(2023, 4, 9),
        )
        mock_encrypt_behavior.assert_called_once_with(
            temp_dir=(Path("some_dir") / "tmp")
        )
        mock_upload_to_s3.assert_called_once_with(
            temp_dir=(Path("some_dir") / "tmp")
        )
        mock_trigger_pipeline.assert_called_once()

        self.assertEqual(1, 1)


if __name__ == "__main__":
    unittest.main()



import logging
import os
import re
from enum import Enum

import dask.array as da

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class BkgSubtraction:

    class Regexes(Enum):
        tile_xyz_pattern = r"tile_[xX]_\d{4}_[yY]_\d{4}_[zZ]_\d{4}"

    @staticmethod
    def subtract(im: da.Array, bkg_da: da.Array) -> da.Array:
        """
        Subtract the background image from the input image volume.

        Parameters:
            im: 3D raw image volume dask array.
            bkg_im: 2D background image dask array.

        Returns:
            Resulting image volume after background subtraction.
        """
        bkg_da = da.pad(
            bkg_da,
            [(0, im.shape[1] - bkg_da.shape[0]), (0, im.shape[2] - bkg_da.shape[1])],
            mode="edge"
        )
        return da.clip((im + 100) - bkg_da, 100, 2 ** 16 - 1) - 100

    @staticmethod
    def get_bkg_path(tile_path: str, bkg_im_dir: str) -> str:
        """
        Get the path to the corresponding background image from the raw tile path and background image directory.

        Parameters:
            tile_path: path to the raw image volume.
            bkg_im_dir: path to the background image directory.

        Returns:
            the path to the background image corresponding to the raw image.
        """
        m = re.search(BkgSubtraction.Regexes.tile_xyz_pattern.value, tile_path)
        if m is None:
            raise ValueError(f"tile name does not follow convention: {tile_path}")
        return os.path.join(bkg_im_dir, "bkg_" + m.group(0) + ".tiff")









"""Module to test that Processing metadata is processed correctly."""
import datetime
import json
import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, mock_open, patch

from aind_data_schema import (
    Procedures,
    Processing,
    RawDataDescription,
    Subject,
)
from aind_data_schema.data_description import Modality
from aind_data_schema.processing import ProcessName
from requests import ConnectionError, Response

from aind_data_transfer.config_loader.ephys_configuration_loader import (
    EphysJobConfigurationLoader,
)
from aind_data_transfer.transformations.metadata_creation import (
    ProceduresMetadata,
    ProcessingMetadata,
    RawDataDescriptionMetadata,
    SubjectMetadata,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
CONFIGS_DIR = TEST_DIR / "resources" / "test_configs"
METADATA_DIR = TEST_DIR / "resources" / "test_metadata"

with open(METADATA_DIR / "processing.json", "r") as f:
    expected_processing_instance_json = json.load(f)

with open(METADATA_DIR / "data_description.json", "r") as f:
    expected_data_description_instance_json = json.load(f)


class TestProcessingMetadata(unittest.TestCase):
    """Tests methods in ProcessingMetadata class"""

    conf_file_path = CONFIGS_DIR / "ephys_upload_job_test_configs.yml"
    args = ["-c", str(conf_file_path)]

    @patch(
        "aind_data_transfer.config_loader.ephys_configuration_loader."
        "EphysJobConfigurationLoader._get_endpoints"
    )
    def test_create_processing_metadata(self, mock_get_endpoints) -> None:
        """Tests that the processing metadata is created correctly."""

        mock_get_endpoints.return_value = {"codeocean_trigger_capsule": None}

        loaded_configs = EphysJobConfigurationLoader().load_configs(self.args)

        start_date_time = datetime.datetime.fromisoformat(
            "2020-10-20T00:00:00.000+00:00"
        )
        end_date_time = datetime.datetime.fromisoformat(
            "2020-10-20T01:00:00.000+00:00"
        )
        input_location = "some_input_location"
        output_location = "some_output_location"
        code_url = "https://github.com/AllenNeuralDynamics/aind-data-transfer"

        parameters = loaded_configs

        processing_metadata = ProcessingMetadata.from_inputs(
            process_name=ProcessName.EPHYS_PREPROCESSING,
            start_date_time=start_date_time,
            end_date_time=end_date_time,
            input_location=input_location,
            output_location=output_location,
            code_url=code_url,
            parameters=parameters,
        )

        # Hack to get match version to be the same as in the example file
        expected_processing_instance_json["data_processes"][0][
            "version"
        ] = processing_metadata.model_obj["data_processes"][0]["version"]

        expected_processing_instance = Processing.parse_obj(
            expected_processing_instance_json
        )

        self.assertEqual(
            json.loads(expected_processing_instance.json()),
            processing_metadata.model_obj
        )
        self.assertEqual(Processing, processing_metadata._model())
        self.assertEqual(
            "processing.json", processing_metadata.output_filename
        )


class TestSubjectMetadata(unittest.TestCase):
    """Tests methods in SubjectMetadata class"""

    successful_response_message = {
        "message": "Valid Model.",
        "data": {
            "describedBy": "https://github-location.org/subject.py",
            "schema_version": "0.3.0",
            "species": "Mus musculus",
            "subject_id": "632269",
            "sex": "Female",
            "date_of_birth": "2022-05-01",
            "genotype": "Pvalb-IRES-Cre/wt;RCL-somBiPoles_mCerulean-WPRE/wt",
            "mgi_allele_ids": None,
            "background_strain": None,
            "source": None,
            "rrid": None,
            "restrictions": None,
            "breeding_group": (
                "Pvalb-IRES-Cre;RCL-somBiPoles_mCerulean-WPRE(ND)"
            ),
            "maternal_id": "615310",
            "maternal_genotype": "Pvalb-IRES-Cre/wt",
            "paternal_id": "623236",
            "paternal_genotype": "RCL-somBiPoles_mCerulean-WPRE/wt",
            "wellness_reports": None,
            "notes": None,
        },
    }

    multiple_subjects_response = {
        "message": "Multiple Items Found.",
        "data": (
            [
                successful_response_message["data"],
                successful_response_message["data"],
            ]
        ),
    }

    @patch("os.path.isdir")
    @patch("builtins.open", new_callable=mock_open())
    @patch(
        "aind_metadata_service.client.AindMetadataServiceClient.get_subject"
    )
    @patch("logging.info")
    def test_successful_response(
        self,
        mock_log_info: MagicMock,
        mock_api_get: MagicMock,
        mock_open: MagicMock,
        mock_os: MagicMock,
    ) -> None:
        """Tests parsing successful response from metadata service."""

        successful_response = Response()
        successful_response.status_code = 200
        successful_response._content = json.dumps(
            self.successful_response_message
        ).encode("utf-8")

        mock_api_get.return_value = successful_response

        actual_subject = SubjectMetadata.from_service(
            "632269", "http://a-fake-url"
        )
        is_model_valid = actual_subject.validate_obj()
        # Mock writing out to a directory
        mock_os.side_effect = [True, False]
        actual_subject.write_to_json(Path("/some_path/"))
        actual_subject.write_to_json(Path("/some_path/subject2.json"))

        expected_subject = self.successful_response_message["data"]

        mock_log_info.assert_called_once_with("Model is valid.")
        mock_open.assert_has_calls(
            [
                call(Path("/some_path/subject.json"), "w"),
                call().__enter__(),
                call()
                .__enter__()
                .write(json.dumps(expected_subject, indent=3, default=str)),
                call().__exit__(None, None, None),
                call(Path("/some_path/subject2.json"), "w"),
                call().__enter__(),
                call()
                .__enter__()
                .write(json.dumps(expected_subject, indent=3, default=str)),
                call().__exit__(None, None, None),
            ]
        )
        self.assertEqual(expected_subject, actual_subject.model_obj)
        self.assertTrue(is_model_valid)

    @patch("logging.warning")
    @patch(
        "aind_metadata_service.client.AindMetadataServiceClient.get_subject"
    )
    def test_multiple_response_warning(
        self,
        mock_api_get: MagicMock,
        mock_log_warn: MagicMock,
    ) -> None:
        """Tests parsing multiples subjects from metadata service."""

        multiple_response = Response()
        multiple_response.status_code = 300
        multiple_response._content = json.dumps(
            self.multiple_subjects_response
        ).encode("utf-8")

        mock_api_get.return_value = multiple_response

        actual_subject = SubjectMetadata.from_service(
            "632269", "http://a-fake-url"
        ).model_obj

        expected_subject = self.successful_response_message["data"]
        mock_log_warn.assert_called_once_with(
            "SubjectMetadata: Multiple Items Found."
        )
        self.assertEqual(expected_subject, actual_subject)

    @patch("logging.warning")
    @patch(
        "aind_metadata_service.client.AindMetadataServiceClient.get_subject"
    )
    def test_invalid_response_warning(
        self,
        mock_api_get: MagicMock,
        mock_log_warn: MagicMock,
    ) -> None:
        """Tests parsing invalid Subject from metadata service."""

        invalid_response = Response()
        invalid_response.status_code = 406
        msg = self.successful_response_message
        msg["message"] = "Validation Errors: Errors here!"
        invalid_response._content = json.dumps(
            self.successful_response_message
        ).encode("utf-8")

        mock_api_get.return_value = invalid_response

        actual_subject = SubjectMetadata.from_service(
            "632269", "http://a-fake-url"
        ).model_obj

        expected_subject = self.successful_response_message["data"]
        mock_log_warn.assert_called_once_with(
            "SubjectMetadata: Validation Errors: Errors here!"
        )
        self.assertEqual(expected_subject, actual_subject)

    @patch("logging.warning")
    @patch("logging.error")
    @patch(
        "aind_metadata_service.client.AindMetadataServiceClient.get_subject"
    )
    def test_server_response_warning(
        self,
        mock_api_get: MagicMock,
        mock_log_err: MagicMock,
        mock_log_warn: MagicMock,
    ) -> None:
        """Tests parsing server error response from metadata service."""

        err_response = Response()
        err_response.status_code = 500
        err_message = {"message": "Internal Server Error.", "data": None}
        err_response._content = json.dumps(err_message).encode("utf-8")
        mock_api_get.return_value = err_response

        actual_subject = SubjectMetadata.from_service(
            "632269", "http://a-fake-url"
        )
        expected_subject = Subject.construct().dict()
        is_model_valid = actual_subject.validate_obj()

        mock_log_err.assert_called_once_with(
            "SubjectMetadata: Internal Server Error."
        )
        mock_log_warn.assert_called_once_with(
            "Validation Errors: 5 validation errors for Subject\nspecies\n  "
            "field required (type=value_error.missing)\nsubject_id\n  "
            "field required (type=value_error.missing)\nsex\n  "
            "field required (type=value_error.missing)\ndate_of_birth\n  "
            "field required (type=value_error.missing)\ngenotype\n  "
            "field required (type=value_error.missing)"
        )
        self.assertEqual(expected_subject, actual_subject.model_obj)
        self.assertEqual("subject.json", actual_subject.output_filename)
        self.assertFalse(is_model_valid)

    @patch("logging.error")
    @patch(
        "aind_metadata_service.client.AindMetadataServiceClient.get_subject"
    )
    def test_no_response_warning(
        self,
        mock_api_get: MagicMock,
        mock_log_err: MagicMock,
    ) -> None:
        """Tests parsing no response from metadata service."""

        mock_api_get.side_effect = ConnectionError("Unable to connect")

        actual_subject = SubjectMetadata.from_service(
            "632269", "http://a-fake-url"
        ).model_obj
        expected_subject = Subject.construct().dict()

        mock_log_err.assert_called_once_with(
            "SubjectMetadata: An error occurred connecting to metadata "
            "service: Unable to connect"
        )
        self.assertEqual(expected_subject, actual_subject)


class TestDataDescriptionMetadata(unittest.TestCase):
    """Tests methods in DataDescriptionMetadata class"""

    def test_create_data_description_metadata(self) -> None:
        """
        Tests that the data description metadata is created correctly.
        """
        data_description = RawDataDescriptionMetadata.from_inputs(
            name="diSPIM_12345_2022-02-21_16-30-01", modality=[Modality.SPIM]
        )

        expected_data_description_instance = RawDataDescription.parse_obj(
            expected_data_description_instance_json
        )

        self.assertEqual(
            json.loads(expected_data_description_instance.json()),
            data_description.model_obj,
        )
        self.assertEqual(RawDataDescription, data_description._model())
        self.assertEqual(
            "data_description.json", data_description.output_filename
        )


class TestProceduresMetadata(unittest.TestCase):
    """Tests methods in the ProceduresMetadata class."""

    successful_response_message = {
        "message": "Valid Model.",
        "data": {
            "describedBy": (
                "https://raw.githubusercontent.com/AllenNeuralDynamics/"
                "aind-data-schema/main/src/aind_data_schema/procedures.py"
            ),
            "schema_version": "0.6.1",
            "subject_id": "436083",
            "subject_procedures": [
                {
                    "start_date": "2019-01-09",
                    "end_date": "2019-01-09",
                    "experimenter_full_name": "NSB-000",
                    "iacuc_protocol": "1804",
                    "animal_weight_prior": 21.3,
                    "animal_weight_post": 23.0,
                    "weight_unit": "gram",
                    "anaesthesia": {
                        "type": "isoflurane",
                        "duration": None,
                        "level": "Select...",
                    },
                    "notes": None,
                    "injection_materials": None,
                    "recovery_time": None,
                    "injection_duration": 10,
                    "workstation_id": "SWS 6",
                    "instrument_id": "NJ#6",
                    "injection_coordinate_ml": -2.3,
                    "injection_coordinate_ap": -2.3,
                    "injection_coordinate_depth": 2.6,
                    "injection_coordinate_unit": "millimeter",
                    "injection_coordinate_reference": None,
                    "bregma_to_lambda_distance": None,
                    "bregma_to_lambda_unit": "millimeter",
                    "injection_angle": 0.0,
                    "injection_angle_unit": "degree",
                    "injection_hemisphere": "Left",
                    "procedure_type": "Nanoject injection",
                    "injection_volume": 200.0,
                    "injection_volume_unit": "nanoliter",
                    "injection_type": "Nanoject (Pressure)",
                },
                {
                    "start_date": "2019-01-09",
                    "end_date": "2019-01-09",
                    "experimenter_full_name": "NSB-000",
                    "iacuc_protocol": "1804",
                    "animal_weight_prior": 21.3,
                    "animal_weight_post": 23.0,
                    "weight_unit": "gram",
                    "anaesthesia": {"type": "isoflurane", "level": "1.5"},
                    "notes": None,
                    "procedure_type": "Headframe",
                    "headframe_type": "CAM-style",
                    "headframe_part_number": "0160-100-10 Rev A",
                    "headframe_material": None,
                    "well_part_number": None,
                    "well_type": "CAM-style",
                },
                {
                    "start_date": "2019-01-09",
                    "end_date": "2019-01-09",
                    "experimenter_full_name": "NSB-000",
                    "iacuc_protocol": "1804",
                    "animal_weight_prior": 21.3,
                    "animal_weight_post": 23.0,
                    "weight_unit": "gram",
                    "anaesthesia": {"type": "isoflurane", "level": "1.5"},
                    "notes": None,
                    "procedure_type": "Craniotomy",
                    "craniotomy_type": "Visual Cortex",
                    "craniotomy_hemisphere": "Left",
                    "craniotomy_coordinates_ml": 2.8,
                    "craniotomy_coordinates_ap": 1.3,
                    "craniotomy_coordinates_unit": "millimeter",
                    "craniotomy_coordinates_reference": None,
                    "bregma_to_lambda_distance": None,
                    "bregma_to_lambda_unit": "millimeter",
                    "craniotomy_size": 5.0,
                    "craniotomy_size_unit": "millimeter",
                    "implant_part_number": None,
                    "dura_removed": True,
                    "protective_material": None,
                    "workstation_id": "SWS 6",
                    "recovery_time": None,
                },
            ],
            "specimen_procedures": None,
            "notes": None,
        },
    }

    @patch(
        "aind_metadata_service.client.AindMetadataServiceClient.get_procedures"
    )
    def test_procedures_from_service(self, mock_api_get: MagicMock):
        """Tests that the procedures is generated from service call."""

        successful_response = Response()
        successful_response.status_code = 200
        successful_response._content = json.dumps(
            self.successful_response_message
        ).encode("utf-8")

        mock_api_get.return_value = successful_response

        procedures = ProceduresMetadata.from_service(
            "436083", "http://a-fake-url"
        )

        expected_subject = self.successful_response_message["data"]

        self.assertEqual(expected_subject, procedures.model_obj)
        self.assertEqual(Procedures, procedures._model())
        self.assertEqual("procedures.json", procedures.output_filename)


if __name__ == "__main__":
    unittest.main()



import logging
import math
import os
import re
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, List, Tuple

import dask.array as da
import fsspec
import h5py
# Importing this alone doesn't work on HPC
# Must manually override HDF5_PLUGIN_PATH environment variable
# in each Dask worker
import hdf5plugin
import numpy as np
import ujson
import zarr
from kerchunk.tiff import tiff_to_zarr
from numpy import dtype


_LOGGER = logging.getLogger(__name__)


class DataReader(ABC):
    def __init__(self, filepath):
        self.filepath = filepath
        super().__init__()

    @abstractmethod
    def as_dask_array(self, chunks: Any = None) -> da.Array:
        pass

    @abstractmethod
    def as_array(self) -> np.ndarray:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def get_shape(self) -> tuple:
        pass

    @abstractmethod
    def get_chunks(self) -> tuple:
        pass

    @abstractmethod
    def get_itemsize(self) -> int:
        pass

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def get_filepath(self) -> str:
        return self.filepath


class TiffReader(DataReader):
    def __init__(self, filepath: str):
        """
        Class constructor

        Args:
            filepath: the path to the Tiff
            refs_dir: the directory to write the temporary references file used by kerchunk
        """
        filepath = str(filepath)
        super().__init__(filepath)
        refs = tiff_to_zarr(filepath)
        # This file needs to be readable by all nodes
        # in the cluster. I'm writing to the current
        # working directory since the code will likely
        # be run from a location accessible by all nodes.
        self._temp_dir = tempfile.TemporaryDirectory(dir=os.getcwd())
        self._refs_file = os.path.join(
            self._temp_dir.name,
            f"references-{Path(filepath).name}.json"
        )
        if os.path.isfile(self._refs_file):
            raise Exception(f"References file already exists: {self._refs_file}")
        Path(self._refs_file).parent.mkdir(parents=True, exist_ok=True)
        with open(self._refs_file, 'wb') as f:
            f.write(ujson.dumps(refs).encode())
        fs = fsspec.filesystem(
            "reference",
            fo=self._refs_file,
            remote_protocol='file'
        )
        self._arr = zarr.open(fs.get_mapper(""), 'r')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def as_dask_array(self, chunks=None):
        """Return the image stack as a Dask Array

        Args:
            chunks: the chunk shape.

        Returns:
            dask.array.Array
        """
        # For performance it is *critical* the initial dask array construction
        # uses single plane chunks, and to then rechunk with the desired chunks
        a = da.from_array(
            self._arr,
            chunks=(1, self._arr.shape[-2], self._arr.shape[-1])
        )
        if chunks is not None:
            a = a.rechunk(chunks)
        return a

    def as_array(self) -> np.ndarray:
        """
        Get the image stack as a ndarray
        """
        return self._arr[:]

    def as_zarr(self):
        """
        Get the image stack as a zarr Array
        """
        return self._arr

    def get_shape(self) -> tuple:
        """
        Get the shape of the image stack
        """
        return self._arr.shape

    def get_chunks(self) -> tuple:
        """
        Get the chunk shape of the image stack
        """
        return self._arr.chunks

    def get_dtype(self) -> dtype:
        """
        Get the data type of the image stack
        """
        return self._arr.dtype

    def get_itemsize(self) -> int:
        """
        Get the number of bytes per element
        """
        return self.get_dtype().itemsize

    def close(self) -> None:
        """Close the file handle to free resources"""
        self._arr = None
        try:
            self._temp_dir.cleanup()
        except Exception as e:
            _LOGGER.error(f"Error removing references dir: {e}")


class MissingDatasetError(Exception):
    pass


class ImarisReader(DataReader):
    DEFAULT_DATA_PATH = "/DataSet/ResolutionLevel 0/TimePoint 0/Channel 0/Data"

    def __init__(self, filepath):
        super().__init__(filepath)
        self.handle = h5py.File(self.filepath, mode="r")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def as_dask_array(self, data_path: str = DEFAULT_DATA_PATH, chunks=True) -> da.Array:
        """Return the image stack as a Dask Array

        Args:
            data_path: the path to the dataset
            chunks: the chunk shape

        Returns:
            dask.array.Array
        """
        # Shape of the full-res image before 0-padding added by ImarisWriter
        lvl = self._get_res_lvl(data_path)
        real_shape_at_lvl = self._get_shape_at_lvl(lvl)

        # This array is 0-padded to be divisible by the Imaris chunk size in each dimension
        a = da.from_array(self.get_dataset(data_path), chunks=chunks)
        # Crop to the "real" shape at the given resolution
        return a[:real_shape_at_lvl[0], :real_shape_at_lvl[1], :real_shape_at_lvl[2]]

    def as_array(self, data_path=DEFAULT_DATA_PATH) -> np.ndarray:
        """
        Get the image stack as a ndarray
        
        Args:
            data_path: the path to the dataset
        """
        lvl = self._get_res_lvl(data_path)
        real_shape_at_lvl = self._get_shape_at_lvl(lvl)
        return self.get_dataset(data_path)[:real_shape_at_lvl[0], :real_shape_at_lvl[1], :real_shape_at_lvl[2]]

    def get_dataset(self, data_path=DEFAULT_DATA_PATH) -> h5py.Dataset:
        """
        Get the dataset at the specified path

        Args:
            data_path: the path to the dataset

        Returns:
            the dataset
        """
        return self.handle[data_path]

    def get_shape(self) -> tuple:
        """
        Get the shape of the image
        """
        info = self.get_dataset_info()
        return int(info.attrs["Z"].tobytes()), int(info.attrs["Y"].tobytes()), int(info.attrs["X"].tobytes())

    def get_chunks(self, data_path=DEFAULT_DATA_PATH) -> tuple:
        """
        Get the chunk shape of the image

        Args:
            data_path: the path to the dataset
        """
        return self.get_dataset(data_path).chunks

    def get_itemsize(self, data_path=DEFAULT_DATA_PATH) -> int:
        """
        Get the bytes per element of the image

        Args:
            data_path: the path to the dataset
        """
        return self.get_dataset(data_path).dtype.itemsize

    def get_handle(self) -> h5py.File:
        """
        Get the file handle
        """
        return self.handle

    def get_dask_pyramid(
            self,
            num_levels: int,
            timepoint: int = 0,
            channel: int = 0,
            chunks: Any = True
    ) -> List[da.Array]:
        """
        Get a multiscale image pyramid as a list of dask arrays.
        The arrays are ordered from highest resolution to lowest.

        Args:
            num_levels: the number of resolution levels to get
            timepoint: the timepoint of the Imaris dataset
            channel: the channel of the Imaris dataset
            chunks: the chunk shape to use for each array.
                    If chunks are larger than the image shape in any dimension,
                    the chunks are clipped to the image shape in that dimension.

        Returns:
            the image pyramid as a list of dask arrays
        """
        darrays = []
        for lvl in range(0, num_levels):
            ds_path = f"/DataSet/ResolutionLevel {lvl}/TimePoint {timepoint}/Channel {channel}/Data"
            if ds_path not in self.get_handle():
                raise MissingDatasetError(f"{ds_path} does not exist")

            if isinstance(chunks, bool) or chunks is None:
                lvl_chunks = chunks
            else:
                lvl_shape = self._get_shape_at_lvl(lvl)
                assert len(chunks) == len(lvl_shape)
                lvl_chunks = list(chunks)
                for i in range(len(chunks)):
                    lvl_chunks[i] = min(chunks[i], lvl_shape[i])

            darrays.append(self.as_dask_array(ds_path, chunks=lvl_chunks))

        return darrays

    def get_dataset_info(self) -> h5py.Dataset:
        """
        Get the DataSetInfo for the image
        """
        return self.get_handle()[f"DataSetInfo/Image"]

    def get_origin(self) -> List[float]:
        """
        Get the origin for the image as a [Z,Y,X] coordinate list
        """
        info = self.get_dataset_info()
        x_min = float(info.attrs["ExtMin0"].tobytes())
        y_min = float(info.attrs["ExtMin1"].tobytes())
        z_min = float(info.attrs["ExtMin2"].tobytes())
        return [z_min, y_min, x_min]

    def get_extent(self) -> List[float]:
        """
        Get the extent for the image as a [Z,Y,X] coordinate list
        """
        info = self.get_dataset_info()
        x_max = float(info.attrs["ExtMax0"].tobytes())
        y_max = float(info.attrs["ExtMax1"].tobytes())
        z_max = float(info.attrs["ExtMax2"].tobytes())
        return [z_max, y_max, x_max]

    def get_voxel_size(self) -> Tuple[List[float], str]:
        """
        Get the voxel size of the image as a [Z,Y,X] list and the unit string

        Returns:
            the voxel size list and unit string
        """
        info = self.get_dataset_info()
        extmin = self.get_origin()
        extmax = self.get_extent()
        x = int(info.attrs["X"].tobytes())
        y = int(info.attrs["Y"].tobytes())
        z = int(info.attrs["Z"].tobytes())
        unit = info.attrs["Unit"].tobytes()
        voxsize = [
            (extmax[0] - extmin[0]) / z,
            (extmax[1] - extmin[1]) / y,
            (extmax[2] - extmin[2]) / x,
        ]
        return voxsize, unit

    def close(self) -> None:
        """Close the file handle to free resources"""
        if self.handle is not None:
            self.handle.close()
            self.handle = None

    def _get_res_lvl(self, data_path: str):
        lvl_rgx = r"/ResolutionLevel (\d+)/"
        m = re.search(lvl_rgx, data_path)
        if m is None:
            raise Exception(f"Could not parse resolution level from path {data_path}")
        return int(m.group(1))

    def _get_shape_at_lvl(self, lvl: int):
        shape = self.get_shape()
        return tuple(int(math.ceil(s / 2**lvl)) for s in shape)


class DataReaderFactory:
    VALID_EXTENSIONS = [".tiff", ".tif", ".h5", ".ims"]

    factory = {}

    def __init__(self):
        self.factory[".tif"] = TiffReader
        self.factory[".tiff"] = TiffReader
        self.factory[".h5"] = ImarisReader
        self.factory[".ims"] = ImarisReader

    def get_valid_extensions(self):
        return self.VALID_EXTENSIONS

    def create(self, filepath: str) -> DataReader:
        """
        Create the corresponding DataReader from the filepath

        Args:
            filepath: the path to the image

        Returns:
            the DataReader instance
        """
        _, ext = os.path.splitext(filepath)
        if ext not in self.VALID_EXTENSIONS:
            raise NotImplementedError(f"File type {ext} not supported")
        return self.factory[ext](filepath)



import argparse
import logging
import re

import fsspec
import zarr
from fsspec import AbstractFileSystem
from gcsfs import GCSFileSystem
from ng_link import NgState
from s3fs import S3FileSystem

from aind_data_transfer.util import file_utils

TILE_PATTERN = r"tile_x_\d{4}_y_\d{4}_z_\d{4}_ch_\d+"
CHANNEL_PATTERN = r"ch_(\d+)"


def _map_channel_colors(channels: list) -> dict:
    """
    Map a list of channel names to unique colors. Only up to 4 channels are
    currently supported.

    Args:
        channels: a list of channel names

    Returns:
        a dict mapping channel names to colors
    """
    colors = ["green", "magenta", "cyan", "yellow"]
    if len(channels) > 4:
        raise Exception("Only up to 4 channels supported.")
    color_map = {channels[i]: colors[i] for i in range(len(channels))}
    return color_map


def _parse_dimensions(zarr_path: str) -> dict:
    """
    Given a url to a zarr dataset, create a dictionary of the dimensions
    used to construct the neuroglancer state. Each dimension has a
    voxel_size field and a unit field, except for channel (c'),
    which has an empty string for unit.

    Args:
        zarr_path: the url of the zarr dataset

    Returns:
        a dict of dimensions
    """
    z = zarr.open(zarr_path, "r")
    attrs = dict(z[next(iter(z.keys()))].attrs)
    ms = attrs["multiscales"][0]
    axes = {el["name"]: el for el in ms["axes"]}
    full_res = [ds for ds in ms["datasets"] if ds["path"] == "0"][0]
    scale = [
        ct
        for ct in full_res["coordinateTransformations"]
        if ct["type"] == "scale"
    ][0]
    voxel_size = scale["scale"]
    dimensions = {
        "x": {"voxel_size": voxel_size[4], "unit": axes["x"]["unit"]},
        "y": {"voxel_size": voxel_size[3], "unit": axes["y"]["unit"]},
        "z": {"voxel_size": voxel_size[2], "unit": axes["z"]["unit"]},
        "c'": {"voxel_size": voxel_size[1], "unit": ""},
        "t": {"voxel_size": voxel_size[0], "unit": axes["t"]["unit"]},
    }
    return dimensions


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str, help="Path to the dataset")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="directory to output the process_output.json file",
    )
    parser.add_argument(
        "--vmin", type=float, default=0, help="display range minimum"
    )
    parser.add_argument(
        "--vmax", type=float, default=500.0, help="display range maximum"
    )
    return parser.parse_args()


def _get_object_path_key(fs: AbstractFileSystem) -> str:
    """
    Given a fsspec AbstractFileSystem, return the key used to access the
    object path

    Args:
        fs: the AbstractFileSystem instance

    Returns
        the key used to access an object path
    """
    if isinstance(fs, S3FileSystem):
        return "Key"
    elif isinstance(fs, GCSFileSystem):
        return "name"
    else:
        raise NotImplementedError


def write_json_from_zarr(
    input_zarr: str, out_json_dir: str, vmin: float, vmax: float
) -> None:
    """
    Create a "process_output.json" file with a neuroglancer link to the
    zarr data. The link is found in the "ng_link" field.

    Args:
        input_zarr: the url of the zarr dataset
        out_json_dir: the directory to write the "process_output.json" file
        vmin: the default minimum of the neuroglancer display range
        vmax: the default maximum of the neuroglancer display range
    """
    provider, bucket, cloud_path = file_utils.parse_cloud_url(input_zarr)
    protocol = provider.replace("://", "")

    fs = fsspec.filesystem(protocol)

    dimensions = _parse_dimensions(input_zarr)

    channels = []
    source_paths = []
    for f in fs.listdir(input_zarr):
        tile_path_field = _get_object_path_key(fs)
        m = re.search(TILE_PATTERN, f[tile_path_field])
        if m is None:
            continue
        tile_source = provider + f[tile_path_field]
        source_paths.append(tile_source)
        tile_name = m.group(0)
        m = re.search(CHANNEL_PATTERN, tile_name)
        channel = int(m.group(1))
        channels.append(channel)

    unique_channels = set(channels)
    color_map = _map_channel_colors(list(unique_channels))

    layers = []
    for source, ch in zip(source_paths, channels):
        layer = {
            "source": source,
            "tab": "rendering",
            "blend": "additive",
            "shader": {
                "color": color_map[ch],
                "emitter": "RGB",
                "vec": "vec3",
            },
            "shaderControls": {  # Optional
                "normalized": {"range": [vmin, vmax]}
            },
        }
        layers.append(layer)

    state = {"dimensions": dimensions, "layers": layers}

    neuroglancer_link = NgState(
        input_config=state,
        mount_service=protocol,
        bucket_path=bucket,
        output_json=out_json_dir,
    )

    logging.info(f"Created link: {neuroglancer_link.get_url_link()}")
    neuroglancer_link.save_state_as_json()


def main():
    args = parse_args()
    write_json_from_zarr(args.input, args.output, args.vmin, args.vmax)


if __name__ == "__main__":
    main()



"""Tests methods in ecephys_config module"""
import os
import unittest
from pathlib import Path

from aind_data_schema.data_description import Modality
from aind_data_schema.processing import ProcessName

from aind_data_transfer.readers.ephys_readers import DataReader
from aind_data_transfer.transformations.ephys_compressors import (
    CompressorName,
    EcephysCompressionParameters,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"
DATA_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream"


class TestEcephysUploadJobConfigs(unittest.TestCase):
    """Tests methods in EcephysUploadJobEndpoints class."""

    def test_defaults(self):
        """Tests defaults are set correctly."""
        ecephys_configs = EcephysCompressionParameters(source=DATA_DIR)

        self.assertEqual(100, ecephys_configs.clip_n_frames)
        self.assertEqual("1s", ecephys_configs.compress_chunk_duration)
        self.assertEqual(
            {"n_jobs": -1}, ecephys_configs.compress_job_save_kwargs
        )
        self.assertEqual(
            150, ecephys_configs.compress_max_windows_filename_len
        )
        self.assertTrue(ecephys_configs.compress_raw_data)
        self.assertEqual("zarr", ecephys_configs.compress_write_output_format)
        self.assertEqual({"level": 3}, ecephys_configs.compressor_kwargs)
        self.assertEqual(
            CompressorName.WAVPACK, ecephys_configs.compressor_name
        )
        self.assertEqual(DataReader.OPENEPHYS, ecephys_configs.data_reader)
        self.assertEqual(DATA_DIR, ecephys_configs.source)
        self.assertEqual(None, ecephys_configs.extra_configs)
        self.assertEqual(Modality.ECEPHYS, ecephys_configs.modality)
        self.assertEqual(
            ProcessName.EPHYS_PREPROCESSING, ecephys_configs.process_name
        )
        self.assertEqual(10000, ecephys_configs.scale_chunk_size)
        self.assertEqual(100, ecephys_configs.scale_num_chunks_per_segment)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pathlib import Path

from parameterized import parameterized

from aind_data_transfer.readers.imaging_readers import ImagingReaders

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_DIR = TEST_DIR / "resources"


class TestImagingReaders(unittest.TestCase):

    imaging_dir = RESOURCES_DIR / "imaging"
    exaspim_dir = imaging_dir / "exaSPIM_125L_2022-08-05_17-25-36"
    mesospim_dir = imaging_dir / "mesoSPIM_125L_2022-08-18_17-05-00"
    dispim_dir = imaging_dir / "diSPIM_12345_2023-03-16_00-00-00"
    generic_dir = imaging_dir / "12345-random-sample"

    @parameterized.expand(
        [
            (ImagingReaders.Readers.exaspim.value, exaspim_dir),
            (ImagingReaders.Readers.mesospim.value, mesospim_dir),
            (ImagingReaders.Readers.dispim.value, dispim_dir),
            (ImagingReaders.Readers.generic.value, generic_dir),
        ]
    )
    def test_get_raw_data_dir(self, reader, input_dir):
        raw_data_dir = ImagingReaders.get_raw_data_dir(reader, input_dir)
        expected_raw_data_dir = input_dir / reader
        self.assertEqual(expected_raw_data_dir, raw_data_dir)

    @parameterized.expand(
        [
            ImagingReaders.Readers.exaspim.value,
            ImagingReaders.Readers.mesospim.value,
            ImagingReaders.Readers.dispim.value,
            ImagingReaders.Readers.generic.value
        ]
    )
    def test_get_raw_data_dir_fails(self, reader):
        with self.assertRaises(FileNotFoundError):
            ImagingReaders.get_raw_data_dir(reader, "Made up directory")


if __name__ == "__main__":
    unittest.main()



import re
from datetime import datetime
from enum import Enum
from typing import List, Optional

from aind_data_schema.base import AindModel
from aind_data_schema.data_description import (
    Funding,
    Institution,
    Modality,
    datetime_from_name_string,
)
from aind_data_schema.device import SizeUnit
from aind_data_schema.imaging.acquisition import AxisName, Immersion
from pydantic import Field

from aind_data_transfer.util import file_utils


class Status(Enum):
    """Dataset status"""

    PENDING = "pending"
    UPLOADING = "uploading"
    UPLOADED = "uploaded"


class DatasetStatus(AindModel):
    """Status of the datasets"""

    status: Status = Field(
        ...,
        description="Status of the dataset on the local storage",
        title="Institution",
        enumNames=[i.value for i in Status],
    )
    # Creating datetime
    status_date = Field(
        datetime.now().date().strftime("%Y-%m-%d"),
        title="Date the flag was created",
    )
    status_time = Field(
        datetime.now().time().strftime("%H-%M-%S"),
        title="Time the flag was created",
    )


class DataDescription(AindModel):
    project: str = Field(..., description="Project name")
    project_id: str = Field(..., description="Project id")


class Acquisition(AindModel):
    experimenter_full_name: str = Field(
        ...,
        description="First and last name of the experimenter.",
        title="Experimenter full name",
    )
    instrument_id: str = Field(..., title="Instrument ID")
    chamber_immersion: Immersion = Field(
        ..., title="Acquisition chamber immersion data"
    )
    sample_immersion: Optional[Immersion] = Field(
        None, title="Acquisition sample immersion data"
    )
    local_storage_directory: Optional[str] = Field(
        None, title="Local storage directory"
    )


class ChannelRegex(Enum):
    CHANNEL_NAME = r"Ex_\d{3}_Em_\d{3}"


class AxisResolution(AindModel):
    axis_name: AxisName = Field(..., title="Axis name")
    unit: SizeUnit = Field(SizeUnit.UM, title="Axis physical units")
    resolution: float = Field(..., title="Axis resolution")


class StitchingParameters(AindModel):
    channel: str = Field(
        ...,
        title="Stitching channel",
        regex=ChannelRegex.CHANNEL_NAME.value,
        description="Channel we want the dataset to be stitched on",
    )
    dataset_resolution: List[AxisResolution] = Field(
        ..., title="Dataset resolution in each axis"
    )


class CCFParameters(AindModel):
    ccf_channels: List[str] = Field(
        ...,
        title="Channels we want to register the dataset on",
        regex=ChannelRegex.CHANNEL_NAME.value,
    )


class CellSegParameters(AindModel):
    cell_seg_channels: List[str] = Field(
        ...,
        title="Channels we want to segment the dataset on",
        regex=ChannelRegex.CHANNEL_NAME.value,
    )


class ProcessingPipeline(AindModel):
    stitching_parameters: StitchingParameters = Field(
        ..., title="Stitching parameters"
    )
    ccf_parameters: Optional[CCFParameters] = Field(
        None, title="Parameters for the CCF registration"
    )
    cell_segmentation_parameters: Optional[CellSegParameters] = Field(
        None, title="Cell segmentation parameters"
    )


class ProcessingManifest(AindModel):
    """Description of the processing manifest file"""

    schema_version: str = Field("0.1.0", title="Schema Version", const=True)
    license: str = Field("CC-BY-4.0", title="License", const=True)

    specimen_id: str = Field(..., title="Specimen ID")
    dataset_status: DatasetStatus = Field(
        ..., title="Dataset status", description="Dataset status"
    )
    institution: Institution = Field(
        ...,
        description="An established society, corporation, foundation or other organization that collected this data",
        title="Institution",
        enumNames=[i.value.name for i in Institution],
    )
    acquisition: Acquisition = Field(
        ...,
        title="Acquisition data",
        description="Acquition data coming from the rig which is necessary to create matadata files",
    )

    processing_pipeline: ProcessingPipeline = Field(
        ...,
        title="SmartSPIM pipeline parameters",
        description="Parameters necessary for the smartspim pipeline parameters",
    )


if __name__ == "__main__":
    # print(ProcessingManifest.schema_json(indent=2))
    # print(ProcessingManifest.schema())

    output_path = "processing_manifest.json"

    processing_manifest_example = ProcessingManifest(
        specimen_id="000000",
        dataset_status=DatasetStatus(status="pending"),
        institution=Institution.AIND,
        acquisition=Acquisition(
            experimenter_full_name="John Rohde",
            instrument_id="SmartSPIM-id-1",
            chamber_immersion=Immersion(
                medium="Cargille oil 1.5200", refractive_index=1.5208
            ),
            sample_immersion=Immersion(
                medium="Cargille oil 1.5200", refractive_index=1.5208
            ),  # Optional parameter
            local_storage_directory="D:/SmartSPIM_Data",
        ),
        processing_pipeline=ProcessingPipeline(
            stitching_parameters=StitchingParameters(
                channel="Ex_488_Em_561",
                dataset_resolution=[
                    AxisResolution(axis_name="X", resolution=1.8),
                    AxisResolution(axis_name="Y", resolution=1.8),
                    AxisResolution(axis_name="Z", resolution=2.0),
                ],
            ),
            ccf_parameters=CCFParameters(
                ccf_channels=["Ex_488_Em_561", "Ex_561_Em_600"]
            ),  # Optional parameter
            cell_segmentation_parameters=CellSegParameters(
                cell_seg_channels=["Ex_488_Em_561", "Ex_561_Em_600"]
            ),  # Optional parameter
        ),
    )

    with open(output_path, "w") as f:
        f.write(processing_manifest_example.json(indent=3))

    # reading back the json
    json_data = file_utils.read_json_as_dict(output_path)
    model = ProcessingManifest(**json_data)
    print(model.json())



import os


class HDF5PluginError(Exception):
    pass


def find_hdf5plugin_path() -> str:
    """
    Finds the site-packages directory containing hdf5plugin plugins. This requires
    that hdf5plugin is installed in your current Python environment.

    Returns:
        the path to the hdf5plugin plugins directory
    """
    # this should work with both conda environments and virtualenv
    # see https://stackoverflow.com/a/46071447
    import sysconfig

    site_packages = sysconfig.get_paths()["purelib"]
    plugin_path = os.path.join(site_packages, "hdf5plugin/plugins")
    if not os.path.isdir(plugin_path):
        raise HDF5PluginError(
            f"Could not find hdf5plugin in site-packages, "
            f"{plugin_path} does not exist. "
            f"Try setting --hdf5_plugin_path manually."
        )
    return plugin_path



import datetime
import json
from pathlib import Path
from typing import Union
import xml.etree.ElementTree as ET

def read_json(json_path: Union[Path, str]) -> dict: 
    with open(json_path) as f:
        return json.load(f)

def read_log_file(log_path: str) -> dict: 
    with open(log_path, 'r') as f: 
        lines = f.readlines()

    log_dict = {}
    log_dict['tiles']: list[dict] = []
    for i, line in enumerate(lines): 
        line = line.replace("\'", "\"")
        tmp = json.loads(f'{line}')
        
        # Only specific lines matter:
        if i == 0: 
            log_dict['session_start_time'] = datetime.datetime.fromisoformat(tmp['created_strftime'])
        if 'local_storage_directory' in tmp.keys():
            tmp.pop('name')
            tmp.pop('msg')
            tmp.pop('levelname')
            tmp.pop('created')
            tmp.pop('created_strftime')
            log_dict = {**log_dict, **tmp}
        if 'file_name' in tmp.keys():
            tmp.pop('name')
            tmp.pop('msg')
            tmp.pop('levelname')
            tmp.pop('created')
            tmp.pop('created_strftime')
            log_dict['tiles'].append(tmp)
        if i == len(lines) - 1: 
            log_dict['session_end_time'] = datetime.datetime.fromisoformat(tmp['created_strftime'])

    return log_dict

def write_xml(tree: ET.ElementTree, path: Union[Path, str]) -> None: 
    ET.indent(tree, space="\t", level=0)
    tree.write(path, encoding="utf-8")


import argparse
import fnmatch
import logging
import os
import time
from pathlib import Path

import tifffile

from aind_data_transfer.transformations.ome_zarr import _compute_chunks
from aind_data_transfer.util.dask_utils import get_client
from aind_data_transfer.util.env_utils import find_hdf5plugin_path
from aind_data_transfer.util.file_utils import any_hdf5, collect_filepaths
from aind_data_transfer.util.io_utils import DataReaderFactory

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
_LOGGER = logging.getLogger(__name__)


_AXES = {"XY": 0, "XZ": 1, "YZ": 2}


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input",
        type=str,
        help="directory of images to transcode",
    )
    parser.add_argument("--output", type=str, help="directory to output MIPs")
    parser.add_argument(
        "--deployment",
        type=str,
        default="local",
        help="cluster deployment type",
    )
    parser.add_argument("--log-level", type=int, default=logging.INFO)
    parser.add_argument(
        "--axes",
        type=str,
        nargs="+",
        default=["XY", "XZ", "YZ"],
        help="the projections to create",
    )
    parser.add_argument(
        "--exclude",
        default=[],
        type=str,
        nargs="+",
        help='filename patterns to exclude, e.g., "*.tif", "*.memento", etc',
    )
    parser.add_argument(
        "--overwrite",
        default=False,
        action="store_true",
        help="Overwrite MIPs if they already exist on disk.",
    )
    parser.add_argument(
        "--chunk-size", type=int, default=128, help="dask chunk size (MB)"
    )
    args = parser.parse_args()
    return args


def project_and_write(arr, axis, out_dir, tile_name, overwrite=False):
    _LOGGER.info(f"computing axis {axis}")
    out_tiff = os.path.join(out_dir, f"{tile_name}_mip_{axis}.tiff")
    if not overwrite and os.path.isfile(out_tiff):
        _LOGGER.info(f"{out_tiff} exists, skipping.")
        return
    norm_axis = _AXES[axis]
    t0 = time.time()
    res = arr.max(axis=norm_axis).compute()
    t1 = time.time()
    _LOGGER.info(f"{t1 - t0}s")
    _LOGGER.info("MIP shape: {res.shape}")
    tifffile.imwrite(out_tiff, res, imagej=True)


def main():
    args = _parse_args()

    _LOGGER.setLevel(args.log_level)

    image_paths = collect_filepaths(
        args.input,
        recursive=True,
        include_exts=DataReaderFactory().VALID_EXTENSIONS,
    )

    exclude_paths = set()
    for path in image_paths:
        if any(fnmatch.fnmatch(path, pattern) for pattern in args.exclude):
            exclude_paths.add(path)

    image_paths = [p for p in image_paths if p not in exclude_paths]

    _LOGGER.info(f"Found {len(image_paths)} images to process")

    if not image_paths:
        _LOGGER.warning("No images found. Exiting.")
        return

    worker_options = {}
    if any_hdf5(image_paths):
        os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
        worker_options["env"] = {
            "HDF5_PLUGIN_PATH": find_hdf5plugin_path(),
            "HDF5_USE_FILE_LOCKING": "FALSE",
        }

    client, _ = get_client(args.deployment, worker_options=worker_options)

    os.makedirs(args.output, exist_ok=True)

    for impath in image_paths:
        _LOGGER.info(f"Processing {impath}")

        reader = DataReaderFactory().create(impath)
        chunks = _compute_chunks(reader, args.chunk_size)[2:]  # MB
        _LOGGER.info(f"chunks: {chunks}")

        arr = reader.as_dask_array(chunks=chunks)

        tile_name = Path(impath).stem

        # FIXME: compute the projections serially since
        #  dask worker memory blows up if doing
        #  xy, xz, yz = dask.compute(arr.max(axis=0), arr.max(axis=1), arr.max(axis=2))
        for axis in args.axes:
            project_and_write(
                arr, axis, args.output, tile_name, overwrite=args.overwrite
            )


if __name__ == "__main__":
    main()



""" Module to handle file transfers to AWS s3 storage.
"""

import logging
import os
from typing import List

from awscrt.s3 import S3Client
from botocore.credentials import create_credential_resolver
from botocore.session import get_session
from s3transfer.constants import GB, MB
from s3transfer.crt import (
    BotocoreCRTRequestSerializer,
    CRTTransferFuture,
    CRTTransferManager,
    create_s3_crt_client,
)

from aind_data_transfer.util.file_utils import (
    collect_filepaths,
    make_cloud_paths,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class S3Uploader:
    """
    Class to handle file and directory uploads to s3
    """

    def __init__(
        self,
        region="us-west-2",
        num_threads=1,
        target_throughput=5 * GB / 8,
        part_size=8 * MB,
        upload_timeout=None,
        use_ssl=True,
        verify=None,
    ):
        session = get_session()
        resolver = create_credential_resolver(session, region_name=region)
        self.s3_crt_client = create_s3_crt_client(
            region,
            botocore_credential_provider=resolver,
            num_threads=num_threads,
            target_throughput=target_throughput,
            part_size=part_size,
            use_ssl=use_ssl,
            verify=verify,
        )
        self.request_serializer = BotocoreCRTRequestSerializer(session)
        self.upload_timeout = upload_timeout

    def upload_file(
        self, filepath: str, s3_bucket: str, s3_key: str
    ) -> List[str]:
        """Upload a single file to s3.
        Args:
            filepath (str): absolute path to file to upload.
            s3_bucket (str): name of s3 bucket
            s3_key (str): location relative to bucket to store blob
        Returns:
            A list of filepaths for failed uploads
        """
        with CRTTransferManager(
            self.s3_crt_client, self.request_serializer
        ) as manager:
            future = manager.upload(filepath, s3_bucket, s3_key)
            return _await_file_upload_futures(
                [future], [filepath], self.upload_timeout
            )

    def upload_files(
        self,
        filepaths: List[str],
        s3_bucket: str,
        s3_folder: str,
        root: str = None,
    ) -> List[str]:
        """Upload a list of files to s3.
        Args:
            filepaths (list): absolute paths of files to upload.
            s3_bucket (str): name of s3 bucket
            s3_folder (str): location relative to bucket to store objects
            root (str): root directory shared by all files in filepaths.
                        If None, all files will be stored as a flat list
                        under s3_folder. Default is None.
        Returns:
            A list of filepaths for failed uploads
        """
        s3_paths = make_cloud_paths(filepaths, s3_folder, root=root)
        with CRTTransferManager(
            self.s3_crt_client, self.request_serializer
        ) as manager:
            futures = []
            for fpath, s3_path in zip(filepaths, s3_paths):
                futures.append(manager.upload(fpath, s3_bucket, s3_path))
            return _await_file_upload_futures(
                futures, filepaths, self.upload_timeout
            )

    def upload_folder(
        self,
        folder: str,
        s3_bucket: str,
        s3_folder: str,
        recursive: bool = True,
        exclude_dirs: list = None,
    ) -> List[str]:
        """Upload a directory to s3.
        Args:
            folder (str): absolute path of the folder to upload.
            s3_bucket (str): name of s3 bucket
            s3_folder (str): location relative to bucket to store objects
            recursive (boolean): upload all sub-directories
        Returns:
            A list of filepaths for failed uploads
        """
        return self.upload_files(
            collect_filepaths(folder, recursive, exclude_dirs=exclude_dirs),
            s3_bucket,
            s3_folder,
            root=folder,
        )

    def get_client(self) -> S3Client:
        return self.s3_crt_client

    def set_client(self, client: S3Client) -> None:
        self.s3_crt_client = client

    def get_timeout(self) -> float:
        return self.upload_timeout

    def set_timeout(self, timeout: float) -> None:
        self.upload_timeout = timeout


def _await_file_upload_futures(
    futures: List[CRTTransferFuture], filepaths: List[str], timeout: float
) -> List[str]:
    n_files = len(filepaths)
    failed_uploads = []
    bytes_uploaded = 0
    for i, (fpath, fut) in enumerate(zip(filepaths, futures)):
        try:
            fut.result(timeout)
            logger.info(f"Uploaded file {i + 1}/{n_files} {fpath}")
            bytes_uploaded += os.stat(fpath).st_size
        except Exception as e:
            logger.error(f"Upload failed for {fpath} \n{e}")
            failed_uploads.append(fpath)
    logger.info(f"{bytes_uploaded / (1024 ** 2)} MiB uploaded")
    return failed_uploads



"""Basic job to upload a data directory to s3 with some metadata attached."""

import argparse
import csv
import logging
import re
import sys
from typing import Any, Dict, List, Optional

from aind_data_transfer.config_loader.base_config import (
    BasicJobEndpoints,
    ModalityConfigs,
)
from aind_data_transfer.jobs.basic_job import BasicJob, BasicUploadJobConfigs


class GenericS3UploadJobList:
    """Class to run multiple jobs defined in a csv file."""

    _MODALITY_ENTRY_PATTERN = re.compile(r"^modality(\d*)$")

    def __init__(self, args: list) -> None:
        """Initializes class with sys args. Convert the sys args to configs."""
        self.args = args
        self.configs = self._load_configs(args)
        self.job_list = self._create_job_config_list()

    @staticmethod
    def _load_configs(args: list) -> argparse.Namespace:
        """Parses args into argparse object."""

        help_message_csv_file = (
            "Path to csv file with list of job configs. The csv file needs "
            "to have the headers: data-source, s3-bucket, subject-id, "
            "modality, acq-date, acq-time. Optional headers: behavior-dir, "
            "metadata-dir."
        )
        help_message_dry_run = (
            "Tests the upload without actually uploading the files."
        )
        help_message_compress_raw_data = (
            "Zip raw data folder before uploading."
        )
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-j",
            "--jobs-csv-file",
            required=True,
            type=str,
            help=help_message_csv_file,
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help=help_message_dry_run,
            default=False,
        )
        parser.add_argument(
            "--compress-raw-data",
            action="store_true",
            help=help_message_compress_raw_data,
            default=None,
        )
        parser.add_argument(
            "--force-cloud-sync",
            action="store_true",
            help="Force syncing of data if s3 location already exists.",
            default=False,
        )
        job_args = parser.parse_args(args)
        return job_args

    @staticmethod
    def _clean_csv_entry(csv_key: str, csv_value: Optional[str]) -> Any:
        """Tries to set the default value for optional settings if the csv
        entry is blank."""
        if (
            csv_value is None or csv_value == "" or csv_value == " "
        ) and BasicUploadJobConfigs.__fields__.get(csv_key) is not None:
            clean_val = BasicUploadJobConfigs.__fields__[csv_key].default
        else:
            clean_val = csv_value.strip()
        return clean_val

    def _map_row_and_key_to_modality_config(  # noqa: C901
        self,
        modality_key: str,
        cleaned_row: Dict[str, Any],
        modality_counts: Dict[str, Optional[int]],
    ) -> Optional[ModalityConfigs]:
        """
        Maps a cleaned csv row and a key for a modality to process into an
        ModalityConfigs object.
        Parameters
        ----------
        modality_key : str
          The column header like modality, modality0, or modality1, etc.
        cleaned_row : Dict[str, Any]
          The csv row that's been cleaned.
        modality_counts : Dict[str, Optional[int]]
          If more than one type of modality is present in the csv row, then
          they will be assigned numerical ids. This will allow multiple of the
          same modalities to be stored under folders like ecephys0, etc.

        Returns
        -------
        Optional[ModalityConfigs]
          None if unable to parse csv row properly.

        """
        modality: str = cleaned_row[modality_key]
        modality_exists = modality is not None and modality != ""
        modality_unindexed = modality_key == "modality" and modality_exists

        # For backwards compatibility, see if configs are set in the old way
        if (
            modality_unindexed
            and cleaned_row.get("modality.source") is not None
        ):
            source = cleaned_row.get("modality.source")
        elif (
            modality_unindexed
            and cleaned_row.get("modality.data_source") is not None
        ):
            source = cleaned_row.get(f"{modality_key}.data_source")
        elif modality_unindexed and cleaned_row.get("source") is not None:
            source = cleaned_row.get("source")
        elif modality_unindexed and cleaned_row.get("data_source") is not None:
            source = cleaned_row.get("data_source")
        elif modality_exists and modality_key != "modality":
            source = cleaned_row.get(f"{modality_key}.source")
        else:
            source = None

        if (
            source is not None
            and modality_unindexed
            and cleaned_row.get("modality.compress_raw_data") is not None
        ):
            compress_raw_data = cleaned_row.get("modality.compress_raw_data")
        elif (
            source is not None
            and modality_unindexed
            and cleaned_row.get("compress_raw_data") is not None
        ):
            compress_raw_data = cleaned_row.get("compress_raw_data")
        elif source is not None and modality_key != "modality":
            compress_raw_data = cleaned_row.get(
                f"{modality_key}.compress_raw_data"
            )
        else:
            compress_raw_data = None

        compress_raw_data = (
            True
            if self.configs.compress_raw_data is True
            else compress_raw_data
        )

        if (
            source is not None
            and modality_unindexed
            and cleaned_row.get("modality.extra_configs") is not None
        ):
            extra_configs = cleaned_row.get("modality.extra_configs")
        elif (
            source is not None
            and modality_unindexed
            and cleaned_row.get("extra_configs") is not None
        ):
            extra_configs = cleaned_row.get("extra_configs")
        elif source is not None and modality_key != "modality":
            extra_configs = cleaned_row.get(f"{modality_key}.extra_configs")
        else:
            extra_configs = None

        if source is None:
            return None
        else:
            modality_configs = ModalityConfigs(
                modality=modality,
                source=source,
                compress_raw_data=compress_raw_data,
                extra_configs=extra_configs,
            )
            num_id = modality_counts.get(modality)
            modality_configs._number_id = num_id
            if num_id is None:
                modality_counts[modality] = 1
            else:
                modality_counts[modality] = num_id + 1
            return modality_configs

    def _parse_modality_configs_from_row(self, cleaned_row: dict) -> None:
        """
        Parses csv row into a list of ModalityConfigs. Will then process the
        cleaned_row dictionary by removing the old modality keys and replacing
        them with just modalities: List[ModalityConfigs.]
        Parameters
        ----------
        cleaned_row : dict
          csv row that contains keys like modality0, modality0.source,
          modality1, modality1.source, etc.

        Returns
        -------
        None
          Modifies cleaned_row dict in-place

        """
        modalities = []
        modality_keys = [
            m
            for m in cleaned_row.keys()
            if self._MODALITY_ENTRY_PATTERN.match(m)
        ]
        modality_counts: Dict[str, Optional[int]] = dict()
        # Check uniqueness of keys
        if len(modality_keys) != len(set(modality_keys)):
            raise KeyError(
                f"Modality keys need to be unique in csv "
                f"header: {modality_keys}"
            )
        for modality_key in modality_keys:
            # modality = cleaned_row[modality_key]
            modality_configs = self._map_row_and_key_to_modality_config(
                modality_key=modality_key,
                cleaned_row=cleaned_row,
                modality_counts=modality_counts,
            )
            if modality_configs is not None:
                modalities.append(modality_configs)

        # Del old modality keys and replace them with list of modality_configs
        for row_key in [
            m
            for m in cleaned_row.keys()
            if (
                m.startswith("modality")
                or m
                in {
                    "data_source",
                    "extra_configs",
                    "source",
                    "compress_raw_data",
                }
            )
        ]:
            del cleaned_row[row_key]
        cleaned_row["modalities"] = modalities

    def _create_job_config_list(self) -> List[BasicJob]:
        """Reads in the csv file and outputs a list of Job Configs."""
        job_list = list()
        param_store_name = None
        job_endpoints = {}
        with open(self.configs.jobs_csv_file, newline="") as csvfile:
            reader = csv.DictReader(csvfile, skipinitialspace=True)
            for row in reader:
                cleaned_row = {
                    k.strip().replace("-", "_"): self._clean_csv_entry(
                        k.strip().replace("-", "_"), v
                    )
                    for k, v in row.items()
                }
                cleaned_row["acq_date"] = BasicUploadJobConfigs.parse_date(
                    cleaned_row["acq_date"]
                )
                cleaned_row["acq_time"] = BasicUploadJobConfigs.parse_time(
                    cleaned_row["acq_time"]
                )
                # Override with flags set in command line
                if self.configs.compress_raw_data is True:
                    cleaned_row["compress_raw_data"] = True
                if self.configs.dry_run is True:
                    cleaned_row["dry_run"] = True
                if self.configs.force_cloud_sync is True:
                    cleaned_row["force_cloud_sync"] = True
                # Avoid downloading endpoints from aws multiple times
                if cleaned_row.get("aws_param_store_name") is not None:
                    # Check if param store is defined in previous row
                    if cleaned_row["aws_param_store_name"] == param_store_name:
                        cleaned_row.update(job_endpoints)
                    # Otherwise, download it from aws
                    else:
                        job_endpoints = BasicJobEndpoints(
                            aws_param_store_name=(
                                cleaned_row["aws_param_store_name"]
                            )
                        ).dict()
                        cleaned_row.update(job_endpoints)
                        param_store_name = cleaned_row["aws_param_store_name"]
                    del cleaned_row["aws_param_store_name"]

                self._parse_modality_configs_from_row(cleaned_row=cleaned_row)

                configs_from_row = BasicUploadJobConfigs(**cleaned_row)
                new_job = BasicJob(job_configs=configs_from_row)
                job_list.append(new_job)
        return job_list

    def run_job(self):
        """Loop through param list and run GenericS3UploadJob"""
        total_jobs = len(self.job_list)
        current_job_num = 1
        logging.info("Starting all jobs...")
        problem_jobs = []
        for one_job in self.job_list:
            logging.info(
                f"Running job {current_job_num} of {total_jobs} "
                f"with params: {one_job.job_configs.dict()}"
            )
            try:
                one_job.run_job()
            except Exception as e:
                logging.error(
                    f"There was a problem processing job {current_job_num} of "
                    f"{total_jobs}. Skipping for now. "
                    f"Error: {e}"
                )
                problem_jobs.append(
                    f"There was a problem processing job {current_job_num} of "
                    f"{total_jobs}. Error: {e}"
                )
            logging.info(
                f"Finished job {current_job_num} of {total_jobs} "
                f"with params: {one_job.job_configs.dict()}"
            )
            current_job_num += 1
        logging.info("Finished all jobs!")
        if len(problem_jobs) > 0:
            logging.error("There were errors processing the following jobs: ")
            for problem_job in problem_jobs:
                logging.error(problem_job)


if __name__ == "__main__":
    sys_args = sys.argv[1:]
    if ("-j" in sys_args) or ("--jobs-csv-file" in sys_args):
        job = GenericS3UploadJobList(sys_args)
    else:
        configs = BasicUploadJobConfigs.from_args(sys_args)
        # This is here for legacy purposes. In the future, all calls to
        # s3_upload_job will require the jobs to be defined in csv file.
        job = BasicJob(job_configs=configs)
    job.run_job()



"""Tests configurations are set properly"""
import os
import unittest
from pathlib import Path

from numcodecs import Blosc

from aind_data_transfer.config_loader.imaging_configuration_loader import (
    ImagingJobConfigurationLoader,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
CONFIGS_DIR = TEST_DIR / "resources" / "test_configs"


class TestImagingJobConfigs(unittest.TestCase):

    code_repo_url = "https://github.com/AllenNeuralDynamics/aind-data-transfer"

    """Basic config loads test"""

    def test_conf_loads(self):
        """Basic config loads test"""

        raw_data_dir = (
            "tests/resources/imaging/exaSPIM_125L_2022-08-05_17-25-36"
        )

        expected_configs = {
            "jobs": {
                "upload_aux_files": False,
                "transcode": True,
            },
            "endpoints": {
                "raw_data_dir": raw_data_dir,
                "dest_data_dir": (
                    "s3://aind-transfer-test/"
                    "exaSPIM_125L_2022-08-05_17-25-36"
                ),
                "code_repo_location": self.code_repo_url,
            },
            "data": {"name": "imaging"},
            "transcode_job": {
                "compressor": {
                    "compressor_name": "blosc",
                    "kwargs": {
                        "cname": "zstd",
                        "clevel": 1,
                        "shuffle": Blosc.SHUFFLE,
                    },
                },
                "chunk_size": 64,
                "resume": False,
                "n_levels": 8,
                "submit_args": {
                    "conda_activate": (
                        "/allen/programs/aind/workgroups/msma/"
                        "cameron.arshadi/miniconda3/bin/"
                        "activate"
                    ),
                    "conda_env": "aind-data-transfer",
                    "cpus_per_task": 1,
                    "mail_user": "cameron.arshadi@alleninstitute.org",
                    "mem_per_cpu": 3000,
                    "nodes": 8,
                    "ntasks_per_node": 8,
                    "run_parent_dir": (
                        "/home/cameron.arshadi/"
                        "exaSPIM-transcode-jobs/"
                        "exaSPIM_125L_2022-08-05_17-25-36"
                    ),
                    "tmp_space": "8GB",
                    "walltime": "72:00:00",
                    "queue": "aind",
                },
            },
        }
        conf_file_path = (
            CONFIGS_DIR / "imaging" / "transcode_job_test_config.yml"
        )

        args = ["-c", str(conf_file_path)]

        loaded_configs = ImagingJobConfigurationLoader().load_configs(args)
        self.assertEqual(loaded_configs, expected_configs)


if __name__ == "__main__":
    unittest.main()



"""Tests methods in base_config module"""
import json
import os
import unittest
from datetime import date, time
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

from aind_data_schema.data_description import ExperimentType, Modality

from aind_data_transfer.config_loader.base_config import (
    BasicJobEndpoints,
    BasicUploadJobConfigs,
    ModalityConfigs,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"
DATA_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream"
CONFIG_FILE = TEST_DIR / "test_configs" / "ephys_upload_job_test_configs.yml"
BEHAVIOR_DIR = TEST_DIR / "v0.6.x_neuropixels_multiexp_multistream" / "Videos"
METADATA_DIR = TEST_DIR / "test_metadata"


class TestJobEndpointsConfigs(unittest.TestCase):
    """Tests methods in BasicJobEndpoints class."""

    EXAMPLE_ENV_VAR1 = {
        "CODEOCEAN_DOMAIN": "some_domain",
        "CODEOCEAN_TRIGGER_CAPSULE_ID": "some_capsule_id",
        "METADATA_SERVICE_DOMAIN": "some_ms_domain",
        "AIND_DATA_TRANSFER_REPO_LOCATION": "some_dtr_location",
        "VIDEO_ENCRYPTION_PASSWORD": "some_password",
    }

    EXAMPLE_PARAM_STORE_RESPONSE = json.dumps(
        {
            "codeocean_domain": "some_domain",
            "codeocean_trigger_capsule_id": "some_capsule_id",
            "metadata_service_domain": "some_ms_domain",
            "aind_data_transfer_repo_location": "some_dtr_location",
            "video_encryption_password_path": "/path/to/password",
            "codeocean_api_token_path": "/path/to/token",
        }
    )

    @mock.patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @mock.patch("boto3.client")
    def test_resolved_from_env_var(self, mock_boto_client: MagicMock):
        """
        Tests that endpoints can be defined using env vars.
        """
        job_endpoints = BasicJobEndpoints()
        self.assertEqual("some_domain", job_endpoints.codeocean_domain)
        self.assertEqual(
            "some_capsule_id", job_endpoints.codeocean_trigger_capsule_id
        )
        self.assertIsNone(job_endpoints.codeocean_trigger_capsule_version)
        self.assertEqual(
            "some_ms_domain", job_endpoints.metadata_service_domain
        )
        self.assertEqual(
            "some_dtr_location", job_endpoints.aind_data_transfer_repo_location
        )
        self.assertEqual(
            "some_password",
            job_endpoints.video_encryption_password.get_secret_value(),
        )
        self.assertIsNone(job_endpoints.codeocean_api_token)
        self.assertFalse(mock_boto_client.called)

    @mock.patch("boto3.client")
    def test_pull_from_aws(self, mock_client: MagicMock):
        """Tests that endpoints are set correctly from aws param store"""

        mock_client.return_value.get_parameter.side_effect = [
            {"Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}},
            {
                "Parameter": {
                    "Value": json.dumps({"password": "some_password"})
                }
            },
            {
                "Parameter": {
                    "Value": json.dumps(
                        {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                    )
                }
            },
        ]
        job_endpoints = BasicJobEndpoints(
            aws_param_store_name="some_param_store_name"
        )
        self.assertEqual("some_domain", job_endpoints.codeocean_domain)
        self.assertEqual(
            "some_capsule_id", job_endpoints.codeocean_trigger_capsule_id
        )
        self.assertIsNone(job_endpoints.codeocean_trigger_capsule_version)
        self.assertEqual(
            "some_ms_domain", job_endpoints.metadata_service_domain
        )
        self.assertEqual(
            "some_dtr_location", job_endpoints.aind_data_transfer_repo_location
        )
        self.assertEqual(
            "some_password",
            job_endpoints.video_encryption_password.get_secret_value(),
        )
        self.assertEqual(
            "some_token", job_endpoints.codeocean_api_token.get_secret_value()
        )


class TestBasicUploadJobConfigs(unittest.TestCase):
    """Tests that the configs for the basic upload job are set correctly."""

    EXAMPLE_ENV_VAR1 = {
        "CODEOCEAN_DOMAIN": "some_domain",
        "CODEOCEAN_TRIGGER_CAPSULE_ID": "some_capsule_id",
        "METADATA_SERVICE_DOMAIN": "some_ms_domain",
        "AIND_DATA_TRANSFER_REPO_LOCATION": "some_dtr_location",
        "VIDEO_ENCRYPTION_PASSWORD": "some_password",
        "S3_BUCKET": "some_bucket",
        "EXPERIMENT_TYPE": "confocal",
        "MODALITIES": f'[{{"modality":"CONFOCAL",'
        f'"source":"{str(DATA_DIR)}"}}]',
        "SUBJECT_ID": "12345",
        "ACQ_DATE": "2020-10-10",
        "ACQ_TIME": "10:10:10",
        "DRY_RUN": "true",
    }

    EXAMPLE_PARAM_STORE_RESPONSE = json.dumps(
        {
            "codeocean_domain": "some_domain",
            "codeocean_trigger_capsule_id": "some_capsule_id",
            "metadata_service_domain": "some_ms_domain",
            "aind_data_transfer_repo_location": "some_dtr_location",
            "video_encryption_password_path": "/path/to/password",
            "codeocean_api_token_path": "/path/to/token",
        }
    )

    @mock.patch.dict(os.environ, EXAMPLE_ENV_VAR1, clear=True)
    @mock.patch("boto3.client")
    def test_resolved_from_env_vars(self, mock_boto_client: MagicMock):
        """Tests that the configs can be set from env vars"""

        basic_job_configs = BasicUploadJobConfigs()
        self.assertEqual("some_domain", basic_job_configs.codeocean_domain)
        self.assertEqual(
            "some_capsule_id", basic_job_configs.codeocean_trigger_capsule_id
        )
        self.assertEqual(
            "some_ms_domain", basic_job_configs.metadata_service_domain
        )
        self.assertEqual(
            "some_dtr_location",
            basic_job_configs.aind_data_transfer_repo_location,
        )
        self.assertEqual(
            "some_password",
            basic_job_configs.video_encryption_password.get_secret_value(),
        )
        self.assertEqual("some_bucket", basic_job_configs.s3_bucket)
        self.assertEqual(
            ExperimentType.CONFOCAL, basic_job_configs.experiment_type
        )
        self.assertEqual(
            [ModalityConfigs(modality=Modality.CONFOCAL, source=DATA_DIR)],
            basic_job_configs.modalities,
        )
        self.assertEqual("12345", basic_job_configs.subject_id)
        self.assertEqual(
            date.fromisoformat("2020-10-10"),
            basic_job_configs.acq_date,
        )
        self.assertEqual(
            time.fromisoformat("10:10:10"),
            basic_job_configs.acq_time,
        )
        self.assertTrue(basic_job_configs.dry_run)
        self.assertIsNone(basic_job_configs.behavior_dir)
        self.assertIsNone(basic_job_configs.metadata_dir)
        self.assertFalse(basic_job_configs.metadata_dir_force)
        self.assertEqual(
            "confocal_12345_2020-10-10_10-10-10", basic_job_configs.s3_prefix
        )
        self.assertFalse(mock_boto_client.called)

    @mock.patch("boto3.client")
    def test_from_req_args(self, mock_client: MagicMock):
        """Tests that the required configs can be set from aws param store"""
        test_req_args = [
            "-b",
            "some_bucket",
            "-s",
            "12345",
            "-e",
            "SmartSPIM",
            "-m",
            f'[{{"modality":"OPHYS","source":"{str(DATA_DIR)}"}}]',
            "-a",
            "2022-10-10",
            "-t",
            "13-24-01",
            "-p",
            "/aws/param/store",
        ]

        mock_client.return_value.get_parameter.side_effect = [
            {"Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}},
            {
                "Parameter": {
                    "Value": json.dumps({"password": "some_password"})
                }
            },
            {
                "Parameter": {
                    "Value": json.dumps(
                        {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                    )
                }
            },
        ]

        basic_job_configs = BasicUploadJobConfigs.from_args(test_req_args)
        self.assertEqual("some_domain", basic_job_configs.codeocean_domain)
        self.assertEqual(
            "some_capsule_id", basic_job_configs.codeocean_trigger_capsule_id
        )
        self.assertEqual(
            "some_ms_domain", basic_job_configs.metadata_service_domain
        )
        self.assertEqual(
            "some_dtr_location",
            basic_job_configs.aind_data_transfer_repo_location,
        )
        self.assertEqual(
            "some_password",
            basic_job_configs.video_encryption_password.get_secret_value(),
        )
        self.assertEqual("some_bucket", basic_job_configs.s3_bucket)
        self.assertEqual(
            ExperimentType.SMARTSPIM, basic_job_configs.experiment_type
        )
        self.assertEqual(
            [ModalityConfigs(modality=Modality.OPHYS, source=DATA_DIR)],
            basic_job_configs.modalities,
        )
        self.assertEqual("12345", basic_job_configs.subject_id)
        self.assertEqual(
            date.fromisoformat("2022-10-10"),
            basic_job_configs.acq_date,
        )
        self.assertEqual(
            time.fromisoformat("13:24:01"),
            basic_job_configs.acq_time,
        )
        self.assertEqual(
            "SmartSPIM_12345_2022-10-10_13-24-01", basic_job_configs.s3_prefix
        )
        self.assertEqual("WARNING", basic_job_configs.log_level)
        self.assertFalse(basic_job_configs.dry_run)
        self.assertIsNone(basic_job_configs.behavior_dir)
        self.assertIsNone(basic_job_configs.metadata_dir)

    @mock.patch("boto3.client")
    def test_from_opt_args(self, mock_client: MagicMock):
        """Tests that the optional configs can also be set."""

        test_opt_args = [
            "-b",
            "some_bucket",
            "-s",
            "12345",
            "-e",
            "SmartSPIM",
            "-m",
            f'[{{"modality":"OPHYS","source":"{str(DATA_DIR)}",'
            f'"extra_configs":"{str(CONFIG_FILE)}"}}]',
            "-l",
            "INFO",
            "-a",
            "2022-10-10",
            "-t",
            "13-24-01",
            "-p",
            "/aws/param/store",
            "-n",
            str(TEST_DIR),
            "-v",
            str(BEHAVIOR_DIR),
            "-x",
            str(METADATA_DIR),
            "--dry-run",
            "--metadata-dir-force",
            "--force-cloud-sync",
        ]

        mock_client.return_value.get_parameter.side_effect = [
            {"Parameter": {"Value": self.EXAMPLE_PARAM_STORE_RESPONSE}},
            {
                "Parameter": {
                    "Value": json.dumps({"password": "some_password"})
                }
            },
            {
                "Parameter": {
                    "Value": json.dumps(
                        {"CODEOCEAN_READWRITE_TOKEN": "some_token"}
                    )
                }
            },
        ]

        basic_job_configs = BasicUploadJobConfigs.from_args(test_opt_args)
        self.assertEqual("some_domain", basic_job_configs.codeocean_domain)
        self.assertEqual(
            "some_capsule_id", basic_job_configs.codeocean_trigger_capsule_id
        )
        self.assertEqual(
            "some_ms_domain", basic_job_configs.metadata_service_domain
        )
        self.assertEqual(
            "some_dtr_location",
            basic_job_configs.aind_data_transfer_repo_location,
        )
        self.assertEqual(
            "some_password",
            basic_job_configs.video_encryption_password.get_secret_value(),
        )
        self.assertEqual("some_bucket", basic_job_configs.s3_bucket)
        self.assertEqual(
            ExperimentType.SMARTSPIM, basic_job_configs.experiment_type
        )
        self.assertEqual(
            [
                ModalityConfigs(
                    modality=Modality.OPHYS,
                    source=DATA_DIR,
                    extra_configs=CONFIG_FILE,
                )
            ],
            basic_job_configs.modalities,
        )
        self.assertEqual("12345", basic_job_configs.subject_id)
        self.assertEqual(
            date.fromisoformat("2022-10-10"),
            basic_job_configs.acq_date,
        )
        self.assertEqual(
            time.fromisoformat("13:24:01"),
            basic_job_configs.acq_time,
        )
        self.assertEqual(
            "SmartSPIM_12345_2022-10-10_13-24-01", basic_job_configs.s3_prefix
        )
        self.assertEqual("INFO", basic_job_configs.log_level)
        self.assertTrue(basic_job_configs.dry_run)
        self.assertTrue(basic_job_configs.metadata_dir_force)
        self.assertTrue(basic_job_configs.force_cloud_sync)
        self.assertEqual(BEHAVIOR_DIR, basic_job_configs.behavior_dir)
        self.assertEqual(METADATA_DIR, basic_job_configs.metadata_dir)

    @mock.patch("boto3.client")
    def test_from_custom_endpoints(self, mock_client: MagicMock):
        """Tests that the endpoints can be defined using a json encoded str"""

        custom_endpoints = {
            "codeocean_domain": "some_domain",
            "codeocean_trigger_capsule_id": "some_capsule_id",
            "metadata_service_domain": "some_ms_domain",
            "aind_data_transfer_repo_location": "some_dtr_location",
            "video_encryption_password": "some_password",
            "codeocean_api_token": "some_token",
        }

        test_req_args = [
            "-b",
            "some_bucket",
            "-s",
            "12345",
            "-e",
            "SmartSPIM",
            "-m",
            f'[{{"modality":"CONFOCAL","source":"{str(DATA_DIR)}"}}]',
            "-a",
            "10/10/2022",
            "-t",
            "13:24:01",
            "-p",
            json.dumps(custom_endpoints),
        ]

        test_malformed_date_args = [
            "-b",
            "some_bucket",
            "-s",
            "12345",
            "-e",
            "SmartSPIM",
            "-m",
            f'[{{"modality":"CONFOCAL","source":"{str(DATA_DIR)}"}}]',
            "-a",
            "12/12/20225",
            "-t",
            "13:24:01",
            "-p",
            json.dumps(custom_endpoints),
        ]

        test_malformed_time_args = [
            "-b",
            "some_bucket",
            "-s",
            "12345",
            "-e",
            "SmartSPIM",
            "-m",
            f'[{{"modality":"CONFOCAL","source":"{str(DATA_DIR)}"}}]',
            "-a",
            "10/10/2022",
            "-t",
            "121:20:20",
            "-p",
            json.dumps(custom_endpoints),
        ]

        test_custom_capsule_args = [
            "-b",
            "some_bucket",
            "-s",
            "12345",
            "-e",
            "SmartSPIM",
            "-m",
            f'[{{"modality":"CONFOCAL","source":"{str(DATA_DIR)}"}}]',
            "-a",
            "10/10/2022",
            "-t",
            "13:24:01",
            "-i",
            "xyz-456",
            "-p",
            json.dumps(custom_endpoints),
        ]

        basic_job_configs = BasicUploadJobConfigs.from_args(test_req_args)
        self.assertEqual("some_domain", basic_job_configs.codeocean_domain)
        self.assertEqual(
            "some_capsule_id", basic_job_configs.codeocean_trigger_capsule_id
        )
        self.assertEqual(
            "some_ms_domain", basic_job_configs.metadata_service_domain
        )
        self.assertEqual(
            "some_dtr_location",
            basic_job_configs.aind_data_transfer_repo_location,
        )
        self.assertEqual(
            "some_password",
            basic_job_configs.video_encryption_password.get_secret_value(),
        )
        self.assertEqual("some_bucket", basic_job_configs.s3_bucket)
        self.assertEqual(
            ExperimentType.SMARTSPIM, basic_job_configs.experiment_type
        )
        self.assertEqual(
            [ModalityConfigs(modality=Modality.CONFOCAL, source=DATA_DIR)],
            basic_job_configs.modalities,
        )
        self.assertEqual("12345", basic_job_configs.subject_id)
        self.assertEqual(
            date.fromisoformat("2022-10-10"),
            basic_job_configs.acq_date,
        )
        self.assertEqual(
            time.fromisoformat("13:24:01"),
            basic_job_configs.acq_time,
        )
        self.assertEqual(
            "SmartSPIM_12345_2022-10-10_13-24-01", basic_job_configs.s3_prefix
        )
        self.assertFalse(basic_job_configs.dry_run)
        self.assertIsNone(basic_job_configs.behavior_dir)
        self.assertIsNone(basic_job_configs.metadata_dir)
        self.assertEqual("WARNING", basic_job_configs.log_level)
        self.assertFalse(mock_client.called)

        with self.assertRaises(ValueError):
            BasicUploadJobConfigs.from_args(test_malformed_date_args)

        with self.assertRaises(ValueError):
            BasicUploadJobConfigs.from_args(test_malformed_time_args)

        custom_capsule_job_configs = BasicUploadJobConfigs.from_args(
            test_custom_capsule_args
        )
        self.assertEqual(
            "xyz-456", custom_capsule_job_configs.codeocean_process_capsule_id
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from aind_data_transfer.s3 import S3Uploader


class TestS3Uploader(unittest.TestCase):
    """Tests methods defined in S3Uploader class"""

    def test_s3_uploader_init(self):
        """Tests that the S3Uploader can be instantiated."""
        uploader = S3Uploader()
        self.assertIsNotNone(uploader)


if __name__ == "__main__":
    unittest.main()



"""This module will have classes that handle mapping to metadata files."""
import json
import logging
import os
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple, Type

import aind_data_schema.base
from aind_data_schema.data_description import (
    Funding,
    Institution,
    Modality,
    RawDataDescription,
)
from aind_data_schema.procedures import Procedures
from aind_data_schema.processing import DataProcess, Processing, ProcessName
from aind_data_schema.subject import Subject
from aind_metadata_service.client import AindMetadataServiceClient
from pydantic import validate_model
from requests import Response
from requests.exceptions import ConnectionError, JSONDecodeError

from aind_data_transfer import __version__ as aind_data_transfer_version
from aind_data_transfer.config_loader.base_config import ModalityConfigs


class MetadataCreation(ABC):
    """Abstract class with convenient methods to handle metadata creation."""

    def __init__(self, model_obj: Optional[dict] = None):
        """
        Default class constructor.
        Parameters
        ----------
        model_obj : Optional[dict]
          The metadata as a dict object. We'll use this representation to
          avoid potential dependency conflicts from aind_data_schema. We can
          validate any of the incoming data using the version of
          aind_data_schema attached to aind_data_transfer.
        """
        self.model_obj = model_obj

    @staticmethod
    @abstractmethod
    def _model() -> Type[aind_data_schema.base.AindCoreModel]:
        """
        Returns
        -------
          Needs to return an AindCoreModel, such as Subject for example.
        """

    @property
    def output_filename(self):
        """Returns the default json file name for the model as defined in
        aind_data_schema."""
        return self._model().construct().default_filename()

    @classmethod
    def from_file(cls, file_location: Path):
        """
        Construct a MetadataCreation instance from a json file.
        Parameters
        ----------
        file_location : Path
          Location of the json file

        """
        with open(file_location) as f:
            contents = json.load(f)
        return cls(model_obj=contents)

    def validate_obj(self) -> bool:
        """
        Validate the model object. Logs a warning if the model_obj is not
        valid.
        Returns
        -------
        bool
          True if the model is valid. False otherwise.

        """
        *_, validation_error = validate_model(self._model(), self.model_obj)
        if validation_error:
            logging.warning(f"Validation Errors: {validation_error}")
            return False
        else:
            logging.info("Model is valid.")
            return True

    def write_to_json(self, path: Path) -> None:
        """
        Write the model_obj to a json file. If the Path is a directory, it will
        use the output_filename method to generate the filename.
        Parameters
        ----------
        path : Path
          Location of where to save the json file. Can be a directory or file.

        Returns
        -------
        None

        """

        if os.path.isdir(path):
            out_path = path / self.output_filename
        else:
            out_path = path

        with open(out_path, "w") as outfile:
            outfile.write(json.dumps(self.model_obj, indent=3, default=str))


class ServiceMetadataCreation(MetadataCreation):
    """Abstract class for metadata pulled from aind_metadata_service"""

    def __init__(self, model_obj: Optional[dict] = None):
        """
        Default class constructor.
        Parameters
        ----------
        model_obj : Optional[dict]
          The metadata as a dict object. We'll use this representation to
          avoid potential dependency conflicts from aind_data_schema. We can
          validate any of the incoming data using the version of
          aind_data_schema attached to aind_data_transfer.
        """
        super().__init__(model_obj=model_obj)

    @staticmethod
    @abstractmethod
    def _get_service_response(
        subject_id: str,
        ams_client: AindMetadataServiceClient,
    ) -> Response:
        """
        Abstract method to retrieve metadata from the service.
        Parameters
        ----------
        subject_id : str
          ID of the subject we want to get metadata for.
        ams_client : AindMetadataServiceClient
          A client to connect to aind_metadata_service.

        Returns
        -------
        Response
          Has a status code and json contents.

        """

    @classmethod
    def from_service(cls, subject_id: str, domain: str):
        """
        Build the class from data pulled from aind_metadata_service.
        Parameters
        ----------
        subject_id : str
          ID of the subject we want to get metadata for.
        domain : str
          Domain name for aind_metadata_service

        """
        ams_client = AindMetadataServiceClient(domain=domain)
        try:
            response = cls._get_service_response(
                subject_id=subject_id, ams_client=ams_client
            )
            response_json = response.json()
            status_code = response.status_code
            # Everything is okay
            if status_code == 200:
                contents = response_json["data"]
            # Multiple items were found
            elif status_code == 300:
                logging.warning(f"{cls.__name__}: {response_json['message']}")
                contents = response_json["data"][0]
            # The data retrieved is invalid
            elif status_code == 406:
                logging.warning(f"{cls.__name__}: {response_json['message']}")
                contents = response_json["data"]
            # Connected to the service, but no data was found
            elif status_code == 404:
                logging.warning(f"{cls.__name__}: {response_json['message']}")
                contents = json.loads(cls._model().construct().json())
            # A serious error happened. Build a default model.
            else:
                logging.error(f"{cls.__name__}: {response_json['message']}")
                contents = json.loads(cls._model().construct().json())
        except (ConnectionError, JSONDecodeError) as e:
            logging.error(
                f"{cls.__name__}: An error occurred connecting to metadata "
                f"service: {e}"
            )
            contents = json.loads(cls._model().construct().json())
        return cls(model_obj=contents)


class SubjectMetadata(ServiceMetadataCreation):
    """Class to manage building the subject metadata"""

    @staticmethod
    def _model() -> Type[aind_data_schema.base.AindCoreModel]:
        """AindDataSchema model"""
        return Subject

    @staticmethod
    def _get_service_response(
        subject_id: str, ams_client: AindMetadataServiceClient
    ) -> Response:
        """
        Method to retrieve metadata from the service.
        Parameters
        ----------
        subject_id : str
          ID of the subject we want to get metadata for.
        ams_client : AindMetadataServiceClient
          A client to connect to aind_metadata_service.

        Returns
        -------
        Response
          Has a status code and json contents.

        """
        return ams_client.get_subject(subject_id)


class ProceduresMetadata(ServiceMetadataCreation):
    """Class to manage building the procedures metadata"""

    @staticmethod
    def _model() -> Type[aind_data_schema.base.AindCoreModel]:
        """AindDataSchema model"""
        return Procedures

    @staticmethod
    def _get_service_response(
        subject_id: str, ams_client: AindMetadataServiceClient
    ) -> Response:
        """
        Method to retrieve metadata from the service.
        Parameters
        ----------
        subject_id : str
          ID of the subject we want to get metadata for.
        ams_client : AindMetadataServiceClient
          A client to connect to aind_metadata_service.

        Returns
        -------
        Response
          Has a status code and json contents.

        """
        return ams_client.get_procedures(subject_id)


class ProcessingMetadata(MetadataCreation):
    """Class to manage building the processing metadata"""

    @staticmethod
    def _model() -> Type[aind_data_schema.base.AindCoreModel]:
        """AindDataSchema model"""
        return Processing

    @classmethod
    def from_inputs(
        cls,
        process_name: ProcessName,
        start_date_time: datetime,
        end_date_time: datetime,
        input_location: str,
        output_location: str,
        code_url: str,
        parameters: dict,
        notes: Optional[str] = None,
    ):
        """
        Build a ProcessingMetadata instance using some basic parameters.
        Parameters
        ----------
        process_name : ProcessName
          Name of the process
        start_date_time : datetime
          Start date and time of the process
        end_date_time : datetime
          End date and time of the process
        input_location : str
          Location of the files that are being processed
        output_location : str
          Location of the files that are being processed
        code_url : str
          Location of the processing code
        parameters : dict
          Parameters used in the process
        notes : Optional[str]
          Optional notes. Defaults to None.

        """
        data_processing_instance = DataProcess(
            name=process_name.value,
            version=aind_data_transfer_version,
            start_date_time=start_date_time,
            end_date_time=end_date_time,
            input_location=input_location,
            output_location=output_location,
            code_url=code_url,
            parameters=parameters,
            notes=notes,
        )
        processing_instance = Processing(
            data_processes=[data_processing_instance]
        )
        # Do this to use enum strings instead of classes in dict representation
        contents = json.loads(processing_instance.json())
        return cls(model_obj=contents)

    @classmethod
    def from_modalities_configs(
        cls,
        modality_configs: List[ModalityConfigs],
        start_date_time: datetime,
        end_date_time: datetime,
        output_location: str,
        code_url: str,
        notes: Optional[str] = None,
    ):
        """
        Build a ProcessingMetadata instance using some basic parameters.
        Parameters
        ----------
        modality_configs : List[ModalityConfigs]
          List of modality configs
        start_date_time : datetime
          Start date and time of the process
        end_date_time : datetime
          End date and time of the process
        output_location : str
          Location of the files that are being processed
        code_url : str
          Location of the processing code
        notes : Optional[str]
          Optional notes. Defaults to None.

        """
        data_processes = []
        for modality_config in modality_configs:
            if modality_config.modality == Modality.ECEPHYS:
                process_name = ProcessName.EPHYS_PREPROCESSING
            else:
                process_name = ProcessName.OTHER
            data_processing_instance = DataProcess(
                name=process_name.value,
                version=aind_data_transfer_version,
                start_date_time=start_date_time,
                end_date_time=end_date_time,
                input_location=str(modality_config.source),
                output_location=output_location,
                code_url=code_url,
                parameters=modality_config.dict(),
                notes=notes,
            )
            data_processes.append(data_processing_instance)
        processing_instance = Processing(data_processes=data_processes)
        # Do this to use enum strings instead of classes in dict representation
        contents = json.loads(processing_instance.json())
        return cls(model_obj=contents)


class RawDataDescriptionMetadata(MetadataCreation):
    """Class to handle the creation of the raw data description metadata
    file."""

    @staticmethod
    def _model() -> Type[aind_data_schema.base.AindCoreModel]:
        """AindDataSchema model"""
        return RawDataDescription

    @classmethod
    def from_inputs(
        cls,
        name: str,
        modality: List[Modality],
        institution: Optional[Institution] = Institution.AIND,
        funding_source: Optional[Tuple] = (
            Funding(funder=Institution.AIND.value.abbreviation),
        ),
        investigators: Optional[List[str]] = None,
    ):
        """
        Build a RawDataDescriptionMetadata instance using some basic
        parameters.
        Parameters
        ----------
        name : str
          Name of the raw data
        modality : List[Modality]
          Modalities of experiment data
        institution : Optional[Institution]
          Primary Institution. Defaults to AIND.
        funding_source : Optional[Tuple]
          Tuple of funding sources. Defaults to (AIND)
        investigators : Optional[List[str]]

        """
        funding_source_list = (
            list(funding_source)
            if isinstance(funding_source, tuple)
            else funding_source
        )
        investigators = [] if investigators is None else investigators
        basic_settings = RawDataDescription.parse_name(name=name)
        data_description_instance = RawDataDescription(
            institution=institution,
            modality=modality,
            funding_source=funding_source_list,
            investigators=investigators,
            **basic_settings,
        )
        # Do this to use enum strings instead of classes in dict representation
        contents = json.loads(data_description_instance.json())
        return cls(model_obj=contents)



"""Job that reads open ephys data, compresses, and writes it."""
import json
import logging
import os
import platform
import subprocess
import sys
import warnings
from datetime import datetime, timezone
from pathlib import Path
from warnings import warn

from aind_codeocean_api.codeocean import CodeOceanClient
from aind_data_schema.processing import ProcessName

from aind_data_transfer.config_loader.ephys_configuration_loader import (
    EphysJobConfigurationLoader,
)
from aind_data_transfer.readers.ephys_readers import EphysReaders
from aind_data_transfer.transformations.ephys_compressors import (
    EphysCompressors,
)
from aind_data_transfer.transformations.metadata_creation import (
    ProceduresMetadata,
    ProcessingMetadata,
    SubjectMetadata,
)
from aind_data_transfer.util.npopto_correction import (
    correct_np_opto_electrode_locations,
)
from aind_data_transfer.util.s3_utils import get_secret
from aind_data_transfer.writers.ephys_writers import EphysWriters

warn(
    f"The module {__name__} is deprecated and will be removed in future "
    f"versions.",
    DeprecationWarning,
    stacklevel=2,
)

root_logger = logging.getLogger()

# Suppress a warning from one of the third-party libraries
deprecation_msg = (
    "Creating a LegacyVersion has been deprecated and will be "
    "removed in the next major release"
)
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, message=deprecation_msg
)


# TODO: Break these up into importable jobs to fix the flake8 warning?
def run_job(args):  # noqa: C901
    # Location of conf file passed in as command line arg
    job_start_time = datetime.now(timezone.utc)
    job_configs = EphysJobConfigurationLoader().load_configs(args)

    root_logger.setLevel(job_configs["logging"]["level"])
    if job_configs["logging"].get("file") is not None:
        fh = logging.FileHandler(job_configs["logging"]["file"])
        fh.setLevel(job_configs["logging"]["level"])
        root_logger.addHandler(fh)

    # Extract raw data name, (e.g., openephys) and raw data path
    data_name = job_configs["data"]["name"]
    data_src_dir = Path(job_configs["endpoints"]["raw_data_dir"])
    dest_data_dir = Path(job_configs["endpoints"]["dest_data_dir"])

    logging.info("Finished loading configs.")

    # Correct NP-opto electrode positions:
    # correction is skipped if Neuropix-PXI version > 0.4.0
    correct_np_opto_electrode_locations(data_src_dir)

    # Clip data job
    if job_configs["jobs"]["clip"]:
        logging.info("Clipping source data. This may take a minute.")
        clipped_data_path = dest_data_dir / "ecephys_clipped"
        clip_kwargs = job_configs["clip_data_job"]["clip_kwargs"]
        streams_to_clip = EphysReaders.get_streams_to_clip(
            data_name, data_src_dir
        )
        aws_secret_names = job_configs["aws_secret_names"]
        secret_name = aws_secret_names.get("video_encryption_password")
        secret_region = aws_secret_names.get("region")
        video_encryption_key_val = {"password": None}
        if secret_name:
            video_encryption_key_val = json.loads(
                get_secret(secret_name, secret_region)
            )

        behavior_directory = job_configs["endpoints"].get("behavior_directory")
        EphysWriters.copy_and_clip_data(
            src_dir=data_src_dir,
            dst_dir=clipped_data_path,
            stream_gen=streams_to_clip,
            behavior_dir=behavior_directory,
            video_encryption_key=video_encryption_key_val["password"],
            **clip_kwargs,
        )

        logging.info("Finished clipping source data.")

    # Compress data job
    if job_configs["jobs"]["compress"]:
        logging.info("Compressing source data.")
        compressed_data_path = dest_data_dir / "ecephys_compressed"
        compressor_name = job_configs["compress_data_job"]["compressor"][
            "compressor_name"
        ]
        compressor_kwargs = job_configs["compress_data_job"]["compressor"][
            "kwargs"
        ]
        format_kwargs = job_configs["compress_data_job"]["format_kwargs"]
        scale_kwargs = job_configs["compress_data_job"]["scale_params"]
        write_kwargs = job_configs["compress_data_job"]["write_kwargs"]
        max_filename_length = job_configs["compress_data_job"].get(
            "max_windows_filename_len"
        )
        read_blocks = EphysReaders.get_read_blocks(data_name, data_src_dir)
        compressor = EphysCompressors.get_compressor(
            compressor_name, **compressor_kwargs
        )
        scaled_read_blocks = EphysCompressors.scale_read_blocks(
            read_blocks, **scale_kwargs
        )
        EphysWriters.compress_and_write_block(
            read_blocks=scaled_read_blocks,
            compressor=compressor,
            output_dir=compressed_data_path,
            max_windows_filename_len=max_filename_length,
            job_kwargs=write_kwargs,
            **format_kwargs,
        )
        logging.info("Finished compressing source data.")

    job_end_time = datetime.now(timezone.utc)
    if job_configs["jobs"]["attach_metadata"]:
        # Processing metadata
        logging.info("Creating processing.json file.")
        start_date_time = job_start_time
        end_date_time = job_end_time
        input_location = data_src_dir
        if job_configs["jobs"]["upload_to_s3"]:
            s3_bucket = job_configs["endpoints"]["s3_bucket"]
            s3_prefix = job_configs["endpoints"]["s3_prefix"]
            aws_dest = f"s3://{s3_bucket}/{s3_prefix}"
            output_location = aws_dest
        elif job_configs["jobs"]["upload_to_gcp"]:
            gcp_bucket = job_configs["endpoints"]["gcp_bucket"]
            gcp_prefix = job_configs["endpoints"]["gcp_prefix"]
            gcp_dest = f"gs://{gcp_bucket}/{gcp_prefix}"
            output_location = gcp_dest
        else:
            output_location = dest_data_dir

        code_url = job_configs["endpoints"]["code_repo_location"]
        parameters = job_configs
        processing_instance = ProcessingMetadata.from_inputs(
            process_name=ProcessName.EPHYS_PREPROCESSING.value,
            start_date_time=start_date_time,
            end_date_time=end_date_time,
            input_location=str(input_location),
            output_location=output_location,
            code_url=code_url,
            parameters=parameters,
            notes=None,
        )
        processing_instance.write_to_json(dest_data_dir)
        logging.info("Finished creating processing.json file.")

        # Subject metadata
        logging.info("Creating subject.json file.")
        metadata_url = job_configs["endpoints"]["metadata_service_url"]
        subject_id = job_configs["data"].get("subject_id")
        if subject_id:
            subject_instance = SubjectMetadata.from_service(
                subject_id=subject_id, domain=metadata_url
            )
        else:
            subject_instance = None
        if subject_instance is not None:
            subject_instance.write_to_json(dest_data_dir)
            logging.info("Finished creating subject.json file.")
        else:
            logging.warning("No subject.json file created!")

        # Procedures metadata
        logging.info("Creating procedures.json file.")
        if subject_id:
            procedures_instance = ProceduresMetadata.from_service(
                subject_id=subject_id, domain=metadata_url
            )
        else:
            procedures_instance = None
        if procedures_instance is not None:
            procedures_instance.write_to_json(dest_data_dir)
            logging.info("Finished creating procedures.json file.")
        else:
            logging.warning("No procedures.json file created!")

    # Upload to s3
    if platform.system() == "Windows":
        shell = True
    else:
        shell = False
    if job_configs["jobs"]["upload_to_s3"]:
        # TODO: Use s3transfer library instead of subprocess?
        logging.info("Uploading to s3.")
        s3_bucket = job_configs["endpoints"]["s3_bucket"]
        s3_prefix = job_configs["endpoints"]["s3_prefix"]
        aws_dest = f"s3://{s3_bucket}/{s3_prefix}"
        if job_configs["upload_data_job"]["dryrun"]:
            subprocess.run(
                [
                    "aws",
                    "s3",
                    "sync",
                    dest_data_dir,
                    aws_dest,
                    "--dryrun",
                ],
                shell=shell,
            )
        else:
            subprocess.run(
                ["aws", "s3", "sync", dest_data_dir, aws_dest], shell=shell
            )
        logging.info("Finished uploading to s3.")

    # Upload to gcp
    if job_configs["jobs"]["upload_to_gcp"]:
        logging.info("Uploading to gcp.")
        gcp_bucket = job_configs["endpoints"]["gcp_bucket"]
        gcp_prefix = job_configs["endpoints"]["gcp_prefix"]
        gcp_dest = f"gs://{gcp_bucket}/{gcp_prefix}"
        if job_configs["upload_data_job"]["dryrun"]:
            subprocess.run(
                [
                    "gsutil",
                    "-m",
                    "rsync",
                    "-r",
                    "-n",
                    dest_data_dir,
                    gcp_dest,
                ],
                shell=shell,
            )
        else:
            subprocess.run(
                ["gsutil", "-m", "rsync", "-r", dest_data_dir, gcp_dest],
                shell=shell,
            )
        logging.info("Finished uploading to gcp.")

    if job_configs["jobs"]["trigger_codeocean_job"]:
        logging.info("Triggering capsule run.")
        capsule_id = job_configs["trigger_codeocean_job"]["capsule_id"]
        co_api_token = os.getenv("CODEOCEAN_API_TOKEN")
        if co_api_token is None:
            aws_secret_names = job_configs["aws_secret_names"]
            secret_name = aws_secret_names.get("code_ocean_api_token_name")
            secret_region = aws_secret_names.get("region")
            token_key_val = json.loads(get_secret(secret_name, secret_region))
            co_api_token = token_key_val["CODEOCEAN_READWRITE_TOKEN"]
        co_domain = job_configs["endpoints"]["codeocean_domain"]
        co_client = CodeOceanClient(domain=co_domain, token=co_api_token)
        run_response = co_client.run_capsule(
            capsule_id=capsule_id,
            data_assets=[],
            parameters=[json.dumps(job_configs)],
        )
        logging.debug(f"Run response: {run_response.json()}")

    return dest_data_dir


if __name__ == "__main__":
    sys_args = sys.argv[1:]
    run_job(sys_args)
